from sympy import *
from sympy import dsolve as sym_dsolve
from functools import reduce
from typing import Union, Optional, List
import warnings
from sympy.core import Symbol, Expr
from sympy.matrices import MatrixBase
import inspect


def _extract_vars(expr: Expr) -> (List[Symbol]):
    """From a sympy expression, extract the free variables sorted by their name.

    Args:
        expr (Expr): Sympy expression to find free variables in.

    Returns:
        List[Symbol]: Sorted list of symbols
    """
    return [e[1] for e in sorted(list(map(lambda x: (str(x), x), expr.free_symbols)))]


def taylor(expr: Expr, vars: List[Symbol], degree: int) -> (Expr):
    """Find the taylor expansion of an expression.

    Args:
        expr (Expr): Sympy expression to do taylor expansion on.
        vars (List[Symbol]): List of symbols to expand on.
        degree (int): Degree of the taylor expansion

    Raises:
        Exception: This function is only defined for one or two variables.

    Returns:
        Expr: The given expression taylor expanded.
    """
    if len(vars) != 2 and len(vars) != 4:
        raise Exception("taylor is only defined for one or two variables")
    if len(vars) == 2:
        return series(expr, vars[0], vars[1], n=degree).removeO()
    P_n = S(0)
    x, x0, y, y0 = vars
    for i in range(degree):
        for j in range(degree - i):
            P_n += (
                expr.diff(x, i).diff(y, j).subs([(x, x0), (y, y0)])
                * (x - x0) ** i
                * (y - y0) ** j
                / (factorial(i) * factorial(j))
            )

    return P_n


def gradient(expr: Expr, var: Optional[List[Symbol]] = None) -> (MatrixBase):
    """Find the gradient of an expression.

    Args:
        expr (Expr): The sympy expression to find the gradient off.
        var (List[Symbol], optional): List of variables to take the gradient with respect to. Defaults to None.

    Returns:
        MatrixBase: The gradient of the expression for the given variables.
    """
    var = _extract_vars(expr) if var is None else var
    if len(var) == 1:
        return expr.diff(var[0])
    return Matrix([expr.diff(v) for v in var])


def hessian(expr: Expr, var: Optional[Symbol] = None) -> (MatrixBase):
    """Find the hessian of an expression.

    Args:
        expr (Expr): The sympy expression to find the hessian off.
        var (list[Symbol], optional): List of variables to find the hessian with respect to. Defaults to estimating them from expr.

    Returns:
        MatrixBase: The hessian of the expression for the given variables.
    """
    var = _extract_vars(expr) if var is None else var
    return Matrix(
        [[reduce(lambda e, s: e.diff(s), [x, y], expr) for x in var] for y in var]
    )


def _extract_field_vars(
    V: MatrixBase, var: Union[List[Symbol], None], namefunc: str
) -> (List[Symbol]):
    """Extract variables used for field (3D) functions.

    Args:
        V (MatrixBase): 3D matrix representing a vector field spanning the variables.
        var (List[Symbol] | None): The 3 variables which the vector field spans. Defaults to estimating them from V.

    Raises:
        AssertionError: Exactly 3 variables (or None) must be specified.

    Returns:
        List[Symbol]: The stated or extracted variables that V spans.
    """

    if var is None:
        var = _extract_vars(V)

        _x, _y, _z = symbols("x,y,z")
        _u, _v, _w = symbols("u,v,w")

        error_str = f"ERROR! Could not guess the variables used. Please specify the variables using {namefunc}(V, var=[x,y,z])."
        warning_str = (
            f"Warning! Variables were not specified. Assuming variables are {var}!"
        )
        assert all(v in [_x, _y, _z, _u, _v, _w] for v in var), error_str

        if all(v in [_x, _y, _z] for v in var):
            var = [_x, _y, _z]
            warnings.warn(warning_str, UserWarning)
        elif all(v in [_u, _v, _w] for v in var):
            var = [_u, _v, _w]
            warnings.warn(warning_str, UserWarning)
        else:
            raise AssertionError(error_str)

    assert len(var) == 3, "Exactly 3 variables must be specified!"

    return var


def div(V: MatrixBase, var: Optional[List[Symbol]] = None) -> (Expr):
    """Find the divergence of a vector field

    Args:
        V (MatrixBase): 3D matrix representing a vector field spanning the variables.
        var (List[Symbol], optional): The 3 variables which the vector field spans. Defaults to estimating them from V.

    Returns:
        Expr: The divergence of the vector field with respect to the given variables.
    """

    var = _extract_field_vars(V, var, "div")

    return sum([V[i].diff(f) for i, f in enumerate(var)])


def rot(V: MatrixBase, var: Optional[List[Symbol]] = None) -> (MatrixBase):
    """Find the rotation of a vector field

    Args:
        V (MatrixBase): 3D matrix representing a vector field spanning the variables.
        var (List[Symbol], optional): The 3 variables which the vector field spans. Defaults to estimating them from V.

    Returns:
        MatrixBase: The rotation of the vector field with respect to the given variables.
    """
    var = _extract_field_vars(V, var, "rot")

    return Matrix(
        [
            V[2].diff(var[1]) - V[1].diff(var[2]),
            V[0].diff(var[2]) - V[2].diff(var[0]),
            V[1].diff(var[0]) - V[0].diff(var[1]),
        ]
    )


def dsolve(ODE: Union[Eq, list, Matrix], ics: Optional[dict] = None) -> (dict):
    """A wrapper for the sympy dsolve-function. Instead of a list of equations denoting the solution, this function returns a dictionary of solutions. This makes it easier to substitute solution in the places it's needed

    Args:
        ODE (Eq, list, Matrix): one or multiple differential equations to be solved by sympy.
        ics (dict, optional): initial/boundary conditions for the differential equation.

    Returns:
        dict: a dictionary containing solutions to 'ODE'
    """

    # Common problem for people is making an single equation, but with both sides
    # being Matrices. This alleviates that issue, and lets users think less
    if (
        type(ODE) == Eq
        and isinstance(ODE.lhs, MatrixBase)
        and isinstance(ODE.rhs, MatrixBase)
    ):
        sol = sym_dsolve(ODE.lhs - ODE.rhs, ics=ics)
    else:
        sol = sym_dsolve(ODE, ics=ics)

    return {eq.lhs: eq.rhs for eq in sol}


def l2_norm(v: MatrixBase) -> (Expr):
    """Computes the l2-norm for a Matrix-class object, without using absolute values on entries, for easier simplification and integration

    Args:
        v (MatrixBase): A sympy Matrix

    Returns:
        Expr: The L2 norm of the matrix
    """
    return sqrt(sum(x**2 for x in v))


def display_equality(lhs: Union[str, Expr], rhs: Union[str, Expr]):
    """Displays two sympy expressions or symbols separated by the equality symbol"
    Args:
        lhs: The left hand side of the equality.
        rhs: The right hand side of the equality.
    Returns:
        Nothing
    """
    if type(lhs) is str:
        lhs =  Symbol(lhs)
    if type(rhs) is str:
        rhs =  Symbol(rhs)
    display(Eq(lhs, rhs))


def display_definition(varname: str, rhs: Optional[Expr] = None):
    """Displays an expression stored in a python variable varname as "varname = expression ...."
    Args:
        varname: The name of the python variable to be displayed on the left hand side.
        rhs: An optional sympy Expr to be displayed on the right hand side instead of the actual variable definition.
    Raises:
        NameError: name 'varname' is not defined.
    Returns:
        Nothing
    """
    if rhs == None:
        caller_vars = inspect.currentframe().f_back.f_locals
        if varname not in caller_vars:
            raise NameError(f"name '{varname}' is not defined")
        rhs = caller_vars[varname]
    display_equality(varname, rhs)



from sympy import *
from dtumathtools import *
import pytest

init_printing()

# remove sympy variable named "test"
test = 0
del test


def test_week1():
    # Store dag
    x, y = symbols("x y")
    f = x * y**2 + x

    res = f.diff(x), f.diff(y)
    assert res == (y**2 + 1, 2 * x * y)

    res = f.diff(x, 2), f.diff(y, 2), f.diff(x, y), f.diff(y, x)
    assert res == (0, 2 * x, 2 * y, 2 * y)

    res = f.diff(x).subs({x: -2, y: 3})
    assert res == 10

    res = f.diff(x, y).subs({x: 5, y: -13})
    assert res == -26

    f = 4 - x**2 - y**2
    p = dtuplot.plot3d(f, (x, -3, 3), (y, -3, 3), show=False)

    dtuplot.plot_contour(f, (x, -3, 3), (y, -3, 3), is_filled=False, show=False)

    zvals = [-2, -1, 0, 1]
    dtuplot.plot_contour(
        f,
        (x, -3, 3),
        (y, -3, 3),
        rendering_kw={"levels": zvals, "alpha": 0.5},
        is_filled=False,
        show=False,
    )

    p = dtuplot.plot3d(f, (x, -3, 3), (y, -3, 3), use_cm=True, legend=True, show=False)

    f = cos(x) + sin(y)
    nf = Matrix([f.diff(x), f.diff(y)])

    dtuplot.plot_vector(
        nf, (x, -pi / 2, 3 / 2 * pi), (y, 0, 2 * pi), scalar=False, show=False
    )

    dtuplot.plot_vector(
        nf,
        (x, -pi / 2, 3 / 2 * pi),
        (y, 0, 2 * pi),
        quiver_kw={"color": "black"},
        contour_kw={"cmap": "Blues_r", "levels": 20},
        grid=False,
        xlabel="x",
        ylabel="y",
        n=15,
        show=False,
    )

    p = dtuplot.plot3d(
        f,
        (x, -pi / 2, 3 / 2 * pi),
        (y, 0, 2 * pi),
        use_cm=True,
        camera={"elev": 45, "azim": -65},
        legend=True,
        show=False,
    )

    f = 1 - x**2 / 2 - y**2 / 2
    x0 = Matrix([1, -1])
    e = Matrix([-1, -2]).normalized()

    p1 = dtuplot.scatter(
        x0,
        rendering_kw={"markersize": 10, "color": "r"},
        xlim=[-2, 2],
        ylim=[-2, 2],
        show=False,
    )
    p1.extend(dtuplot.quiver(x0, e, show=False))

    Nabla = Matrix([diff(f, x), diff(f, y)]).subs({x: x0[0], y: x0[1]})
    a = e.dot(Nabla)
    u, t = symbols("u t", real=True)
    r = Matrix([u * cos(u), u * sin(u)])
    rd = diff(r, u)
    u0 = 3 * pi / 2
    rdu0 = rd.subs(u, u0)
    ru0 = r.subs(u, u0)
    T = ru0 + t * rdu0

    p = dtuplot.plot_parametric(
        r[0],
        r[1],
        (u, 0, 4 * pi),
        rendering_kw={"color": "red"},
        use_cm=False,
        show=False,
    )
    p.extend(
        dtuplot.plot_parametric(
            T[0],
            T[1],
            (t, -1.5, 1.5),
            rendering_kw={"color": "royalblue"},
            use_cm=False,
            show=False,
        )
    )
    p.extend(
        dtuplot.scatter(
            ru0, rendering_kw={"markersize": 10, "color": "black"}, show=False
        )
    )
    p.extend(dtuplot.quiver(ru0, rdu0, rendering_kw={"color": "black"}, show=False))

    # Lille dag
    x, y = symbols("x y")
    g = y**2 - x**2
    res = solveset(Eq(g, 0), y)
    assert res == {-x, x}

    plot_implicit(Eq(g, 0), show=False)

    res = (
        g.subs({x: 1, y: 0}),
        g.subs({x: 0, y: 1}),
        g.subs({x: -1, y: 0}),
        g.subs({x: 0, y: -1}),
    )
    assert res == (-1, 1, -1, 1)

    plot_implicit(g > 0, show=False)

    f = 4 - x**2 - y**2
    dtuplot.plot3d(f, (x, -2, 2), (y, -2, 2), use_cm=True, legend=True, show=False)

    a = {x: 1, y: -1}
    P1 = f.subs(a) + f.diff(x).subs(a) * (x - 1) + f.diff(y).subs(a) * (y + 1)
    assert P1 == -2 * x + 2 * y + 6

    def poly_approx(f, var, expand_from, degree):
        for i, x in enumerate(var):
            f = f.series(x, expand_from[i], degree + 1).removeO()
        return f

    res = poly_approx(f, (x, y), (1, -1), 1)
    assert res == -2 * x + 2 * y + 6

    N = Matrix([2, -2, 1])
    a0 = Matrix([1, -1, f.subs(a)])
    p = dtuplot.plot3d(
        f,
        (x, -2, 2),
        (y, -2, 2),
        rendering_kw={"alpha": 0.3, "cmap": "Blues"},
        xlim=(-2, 3),
        ylim=(-3, 2),
        zlim=(-1, 4),
        camera={"elev": 10, "azim": -110},
        show=False,
    )
    p.extend(
        dtuplot.plot3d(
            P1,
            (x, 0, 2),
            (y, -2, 0),
            rendering_kw={"color": "green", "alpha": 0.5},
            show=False,
        )
    )
    p.extend(
        dtuplot.scatter(
            a0, rendering_kw={"alpha": 1, "s": 100, "color": "black"}, show=False
        )
    )
    p.extend(dtuplot.quiver(a0, N, {"color": "red"}, show=False))

    x, y, u = symbols("x y u")
    f = x**2 / 10 + y**2 / 10 + 10
    r = Matrix([u * cos(u), u * sin(u)])
    h = f.subs({x: r[0], y: r[1]})
    p = dtuplot.plot3d(
        f,
        (x, -10, 10),
        (y, -10, 10),
        use_cm=True,
        camera={"elev": 20, "azim": -125},
        rendering_kw={"alpha": 0.4},
        show=False,
    )
    p.extend(
        dtuplot.plot3d_parametric_line(
            r[0],
            r[1],
            0,
            (u, 0, 4 * pi),
            use_cm=False,
            rendering_kw={"color": "red"},
            show=False,
        )
    )
    r1 = r.subs(u, 3 * pi)
    p.extend(
        dtuplot.scatter(
            r1[0], r1[1], 0, rendering_kw={"color": "black", "s": 100}, show=False
        )
    )
    p.extend(
        dtuplot.quiver(
            r1[0],
            r1[1],
            0,
            0,
            0,
            f.subs({x: r1[0], y: r1[1]}),
            rendering_kw={"color": "orange"},
            show=False,
        )
    )
    r2 = r.subs(u, 3 * pi / 2)
    p.extend(
        dtuplot.scatter(
            r2[0], r2[1], 0, rendering_kw={"color": "black", "s": 100}, show=False
        )
    )
    p.extend(
        dtuplot.quiver(
            r2[0],
            r2[1],
            0,
            0,
            0,
            f.subs({x: r2[0], y: r2[1]}),
            rendering_kw={"color": "orange"},
            show=False,
        )
    )

    dh = h.diff(u)
    dh.subs(u, 3 * pi / 2), dh.subs(u, 3 * pi)
    x, y, z = symbols("x y z")
    f = 1 - x**2 / 2 - y**2 / 2

    p0 = Matrix([1, -1, 0])
    p_parab = dtuplot.plot3d(
        f,
        (x, -2, 2),
        (y, -2, 2),
        use_cm=True,
        camera={"elev": 20, "azim": -80},
        rendering_kw={"alpha": 0.4},
        show=False,
    )
    p_point = dtuplot.scatter(p0, rendering_kw={"color": "black", "s": 100}, show=False)
    res = p_parab + p_point

    dfx = f.diff(x)
    t = symbols("t")
    xtan = p0 + Matrix([1, 0, dfx.subs({x: p0[0], y: p0[1]})]) * t
    n0 = Matrix([0, 1, 0])
    p_xplane = dtuplot.plot_geometry(
        Plane(p0, n0),
        (x, -2, 2),
        (y, -2, 2),
        (z, -3, 1),
        rendering_kw={"alpha": 0.3, "color": "blue"},
        show=False,
    )
    p_xtan = dtuplot.plot3d_parametric_line(
        xtan[0],
        xtan[1],
        xtan[2],
        (t, -1, 1),
        use_cm=False,
        rendering_kw={"color": "red"},
        show=False,
    )
    res = p_parab + p_point + p_xplane + p_xtan

    dfx.subs({x: p0[0], y: p0[1]})
    p_parab2 = dtuplot.plot3d(
        f,
        (x, -2, 2),
        (y, -2, 2),
        use_cm=True,
        camera={"elev": 15, "azim": -60},
        rendering_kw={"alpha": 0.4},
        show=False,
    )
    dfy = f.diff(y)
    ytan = p0 + Matrix([0, 1, dfy.subs({x: p0[0], y: p0[1]})]) * t
    n1 = Matrix([1, 0, 0])
    p_yplane = dtuplot.plot_geometry(
        Plane(p0, n1),
        (x, -2, 2),
        (y, -2, 2),
        (z, -3, 1),
        rendering_kw={"alpha": 0.3, "color": "green"},
        show=False,
    )
    p_ytan = dtuplot.plot3d_parametric_line(
        ytan[0],
        ytan[1],
        ytan[2],
        (t, -1, 1),
        use_cm=False,
        rendering_kw={"color": "red"},
        show=False,
    )
    res = p_parab2 + p_point + p_yplane + p_ytan

    assert dfy.subs({x: p0[0], y: p0[1]}) == 1

    n2 = Matrix([1, 0, dfx]).cross(Matrix([0, 1, dfy]))
    n3 = n2.subs({x: p0[0], y: p0[1]})

    p_xyplane = dtuplot.plot_geometry(
        Plane(p0, n3),
        (x, -2, 2),
        (y, -2, 2),
        (z, -3, 1),
        rendering_kw={"alpha": 0.5, "color": "grey"},
        show=False,
    )
    res = p_parab + p_point + p_xtan + p_ytan + p_xyplane

    e = Matrix([-S(1) / sqrt(5), -S(2) / sqrt(5)])
    p_2dline = dtuplot.plot(
        2 * x - 3,
        (x, -2, 2),
        xlim=(-2, 2),
        ylim=(-2, 2),
        rendering_kw={"color": "grey"},
        show=False,
    )
    p_2darrow = dtuplot.quiver((1, -1), e, rendering_kw={"color": "red"}, show=False)
    p_2dpoint = dtuplot.scatter(
        1, -1, rendering_kw={"color": "black", "s": 10}, show=False
    )
    res = p_2dline + p_2darrow + p_2dpoint

    n4 = Matrix([2, -1, 0])  # normal vektoren tilsvarende y=2x-3
    p_vertical_plane = dtuplot.plot_geometry(
        Plane(p0, n4),
        (x, -2, 2),
        (y, -2, 2),
        (z, -3, 1),
        rendering_kw={"alpha": 0.35, "color": "blue"},
        show=False,
    )
    p_normalvector = dtuplot.quiver(p0, n4, rendering_kw={"color": "red"}, show=False)
    p_arrow_along_plane = dtuplot.quiver(
        p0,
        [e[0], e[1], f.subs({x: 1, y: -1})],
        rendering_kw={"color": "orange"},
        show=False,
    )
    res = p_parab + p_point + p_vertical_plane + p_normalvector + p_arrow_along_plane

    assert Matrix([dfx, dfy]).subs({x: 1, y: -1}) == Matrix([-1, 1])

    a = e.dot(Nabla)
    assert a == -sqrt(5) / 5

    r = Matrix([e[0], e[1], a])
    tangent = p0 + r * t
    p_linetan = dtuplot.plot3d_parametric_line(
        tangent[0],
        tangent[1],
        tangent[2],
        (t, -2, 2),
        use_cm=False,
        rendering_kw={"color": "red"},
        show=False,
    )
    res = p_parab + p_point + p_vertical_plane + p_linetan

    res = (
        p_parab
        + p_point
        + p_linetan
        + p_xyplane
        + p_vertical_plane
        + p_normalvector
        + p_arrow_along_plane
    )


def test_week2():
    # Store dag
    x, y = symbols("x y")
    f = sin(x**2 + y**2)
    dtuplot.plot3d(
        f, (x, -1.5, 1.5), (y, -1.5, 1.5), rendering_kw={"color": "blue"}, show=False
    )
    P1 = dtutools.taylor(f, [x, 0, y, 0], degree=2)
    assert P1 == 0
    p = dtuplot.plot3d(
        P1,
        (x, -1.5, 1.5),
        (y, -1.5, 1.5),
        show=False,
        rendering_kw={"alpha": 0.5},
        camera={"azim": -81, "elev": 18},
    )
    p.extend(dtuplot.plot3d(f, (x, -1.5, 1.5), (y, -1.5, 1.5), show=False))
    P2 = dtutools.taylor(f, [x, 0, y, 0], 3)
    assert P2 == x**2 + y**2
    dtuplot.plot3d(f, P2, (x, -1.5, 1.5), (y, -1.5, 1.5), show=False)
    P6 = dtutools.taylor(f, [x, 0, y, 0], 7)
    p = dtuplot.plot3d(f, (x, -1.5, 1.5), (y, -1.5, 1.5), show=False)
    p.legend = True
    f_p1 = f.subs([(x, 1 / 5), (y, 1 / 5)])
    P1_p1 = P1.subs([(x, 1 / 5), (y, 1 / 5)])
    P2_p1 = P2.subs([(x, 1 / 5), (y, 1 / 5)])
    P6_p1 = P6.subs([(x, 1 / 5), (y, 1 / 5)])
    res = f_p1 - P1_p1, f_p1 - P2_p1, f_p1 - P6_p1
    f_p2 = f.subs([(x, 1 / 2), (y, 1 / 2)])
    P1_p2 = P1.subs([(x, 1 / 2), (y, 1 / 2)])
    P2_p2 = P2.subs([(x, 1 / 2), (y, 1 / 2)])
    P6_p2 = P6.subs([(x, 1 / 2), (y, 1 / 2)])
    res = f_p2 - P1_p2, f_p2 - P2_p2, f_p2 - P6_p2
    z = symbols("z")
    f = (
        7 * x**2
        - 4 * x * y
        + 6 * y**2
        - 4 * y * z
        + 5 * z**2
        - 2 * x
        + 20 * y
        - 10 * z
    )
    H = dtutools.hessian(f)
    var_vec = Matrix([x, y, z])
    A = S(1) / 2 * H
    k = (var_vec.T * A * var_vec)[0].expand()
    assert k == 7 * x**2 - 4 * x * y + 6 * y**2 - 4 * y * z + 5 * z**2
    f_matrix = k + (Matrix([-2, 20, -10]).T * var_vec)[0]

    # Lille dag
    x, y = symbols("x y")
    var_vec = Matrix([x, y])
    lignM = Eq(11 * x**2 - 24 * x * y + 4 * y**2 - 20 * x + 40 * y - 60, 0)
    H = dtutools.hessian(lignM.lhs)
    A = S(1) / 2 * H
    k = (var_vec.T * S(1) / 2 * H * var_vec)[0].expand()
    l = (Matrix([-20, 40]).T * var_vec)[0].expand()
    assert k == 11 * x**2 - 24 * x * y + 4 * y**2
    assert l == -20 * x + 40 * y
    A.eigenvects()
    q1 = Matrix([S(3) / 4, 1])
    q2 = Matrix([-S(4) / 3, 1])
    q1 = q1.normalized()  # Eller q1 = q1 / sqrt((q1.T * q1)[0])
    q2 = q2.normalized()  # q2 = q2 / sqrt((q2.T * q2)[0])
    assert q1 == Matrix([S(3) / 5, S(4) / 5])
    assert q2 == Matrix([-S(4) / 5, S(3) / 5])
    Q = Matrix.hstack(q1, q2)
    assert Q.det() == 1
    Lambda = Matrix([[-5, 0], [0, 20]])
    assert Eq(Lambda, Q.T * A * Q) == True
    x1, y1 = symbols("x1 y1")
    var_vec_1 = Matrix([x1, y1])
    lignM_1 = (
        (var_vec_1.T * Q.T * A * Q * var_vec_1)[0]
        + (Matrix([-20, 40]).T * Q * var_vec_1)[0]
        - 60
    )
    Eq(lignM_1, 0)
    -5 * (x1 - 2) ** 2, expand(-5 * (x1 - 2) ** 2), 20 * (y1 + 1) ** 2, expand(
        20 * (y1 + 1) ** 2
    )
    lign = Eq((-5 * (x1 - 2) ** 2 + 20 * (y1 + 1) ** 2 - 60), 0)
    lign
    lign = Eq((lign.lhs + 60) / 60, (lign.rhs + 60) / 60)
    lign
    assert Q * Matrix([2, -1]) == Matrix([2, 1])
    ((Q * Matrix([S(1) / 2, 1])).T * var_vec)[0], (
        (Q * Matrix([-S(1) / 2, 1])).T * var_vec
    )[0] + 2
    t = symbols("t")
    M = dtuplot.plot_implicit(
        lignM, (x, -4, 8), (y, -6, 6), show=False, aspect="equal", size=(6, 6)
    )
    sym1 = dtuplot.plot_parametric(
        2 + q1[0] * t, 1 + q1[1] * t, (t, -20, 20), show=False
    )
    sym2 = dtuplot.plot_parametric(
        2 + q2[0] * t, 1 + q2[1] * t, (t, -20, 20), show=False
    )
    asymp1 = dtuplot.plot_implicit(
        Eq(-x / S(2) + y, 0), (x, -4, 8), (y, -6, 6), show=False
    )
    asymp2 = dtuplot.plot_implicit(
        Eq(-S(11) * x / 10 + y / S(5) + 2, 0), (x, -4, 8), (y, -6, 6), show=False
    )
    q1p = dtuplot.quiver(
        Matrix([2, 1]), q1 * 2, rendering_kw={"color": "black"}, show=False
    )
    q2p = dtuplot.quiver(
        Matrix([2, 1]), q2 * 2, rendering_kw={"color": "black"}, show=False
    )
    M.extend(sym1)
    M.extend(sym2)
    M.extend(asymp1)
    M.extend(asymp2)
    M.extend(q1p)
    M.extend(q2p)
    M.legend = False


def test_week3():
    x, y = symbols("x y")
    f = x**3 - 3 * x**2 + y**3 - 3 * y**2
    lign1 = Eq(f.diff(x), 0)
    lign2 = Eq(f.diff(y), 0)
    assert f.diff(x) == 3 * x**2 - 6 * x
    assert f.diff(y) == 3 * y**2 - 6 * y
    sols = nonlinsolve([lign1, lign2], [x, y])
    assert (0, 0) in sols
    assert (0, 2) in sols
    assert (2, 0) in sols
    assert (2, 2) in sols
    fxx = f.diff(x, 2)
    fxy = f.diff(x, y)
    fyy = f.diff(y, 2)
    H = Matrix([[fxx, fxy], [fxy, fyy]])
    assert H == Matrix([[(6 * (x - 1)).factor(), 0], [0, (6 * (y - 1)).factor()]])
    [H.subs([(x, x0), (y, y0)]) for (x0, y0) in sols]
    assert dtutools.taylor(f, [x, 0, y, 0], 3) == -3 * x**2 - 3 * y**2
    assert (
        dtutools.taylor(f, [x, 2, y, 2], 3) == 3 * (x - 2) ** 2 + 3 * (y - 2) ** 2 - 8
    )
    pf = dtuplot.plot3d(
        f,
        (x, -0.8, 2.8),
        (y, -0.8, 2.8),
        use_cm=True,
        show=False,
        wireframe=True,
        rendering_kw={"alpha": 0.6},
    )
    # Following command changed, as the scatter function changed. 
    # Each point is now given in a list, so '*' is removed.
    points = dtuplot.scatter(
        [Matrix([x0, y0, f.subs([(x, x0), (y, y0)])]) for (x0, y0) in sols],
        show=False,
        rendering_kw={"s": 100, "color": "red"}
    )
    pf.camera = {"azim": -50, "elev": 30}
    pf.extend(points)
    f = x**4 + 4 * x**2 * y**2 + y**4 - 4 * x**3 - 4 * y**3 + 2
    lign1 = Eq(f.diff(x), 0)
    lign2 = Eq(f.diff(y), 0)
    lign1, lign2, f
    sols = nonlinsolve([lign1, lign2], [x, y])
    assert (0, 0) in sols
    assert (0, 3) in sols
    assert (1, 1) in sols
    assert (3, 0) in sols
    [(N(xsol), N(ysol)) for (xsol, ysol) in sols]
    stat_punkter = list(sols)[:-2]
    H = dtutools.hessian(f)
    assert (H[1, 0]).factor() == 16 * x * y
    Hesse_matricer = [H.subs([(x, x0), (y, y0)]) for x0, y0 in stat_punkter]
    Eig_Hesse_matricer = [h.eigenvals() for h in Hesse_matricer]
    list(zip(Hesse_matricer, Eig_Hesse_matricer))
    dtuplot.plot(f.subs(y, x), 2, (x, -0.5, 1.2), axis_center="auto", show=False)
    tmp = f.subs(y, x) - f.subs([(x, 0), (y, 0)])
    assert tmp == 6 * x**4 - 8 * x**3
    assert dtutools.taylor(f, [x, 0, y, 0], 3) == 2
    pf = dtuplot.plot3d(
        f,
        (x, -2, 5),
        (y, -2, 5),
        xlim=(-2, 5),
        ylim=(-2, 5),
        zlim=(-30, 15),
        show=False,
        rendering_kw={"alpha": 0.5},
    )
    # Following command changed, as the scatter function changed. 
    # Each point is now given in a list, so '*' is removed.
    points = dtuplot.scatter(
        [Matrix([x0, y0, f.subs([(x, x0), (y, y0)])]) for x0, y0 in stat_punkter],
        show=False,
        rendering_kw={"color": "red", "s": 30}
    )
    pf.camera = {"azim": -161, "elev": -5}
    res = pf + points
    with pytest.warns(
        match="The provided expression contains Boolean functions. In order to plot the expression, the algorithm automatically switched to an adaptive sampling."
    ):
        M = dtuplot.plot_implicit(
            Eq(x, 0),
            Eq(y, 1 - x),
            Eq(y, -1),
            (1 - x >= y) & (y >= -1) & (x >= 0),
            (x, -0.1, 2.1),
            (y, -1.1, 1.1),
            show=False,
        )
    M.legend = False
    f = 3 + x - x**2 - y**2
    niveau = dtuplot.plot_contour(f, (x, -0.1, 2.1), (y, -1.1, 1.1), show=False)
    niveau.extend(
        dtuplot.plot_implicit(
            Eq(x, 0),
            Eq(y, 1 - x),
            Eq(y, -1),
            (x, -0.1, 2.1),
            (y, -1.1, 1.1),
            show=False,
        )
    )
    lign1 = Eq(f.diff(x), 0)
    lign2 = Eq(f.diff(y), 0)
    sols = nonlinsolve([lign1, lign2], [x, y])
    assert (S(1) / 2, 0) in sols
    dtuplot.plot(f.subs(y, -1), (x, 0, 2), show=False)
    dtuplot.plot(f.subs(y, 1 - x), (x, 0, 2), show=False)
    dtuplot.plot(f.subs(x, 0), (y, -1, 1), ylim=(0, 3), aspect="equal", show=False)
    stat_punkter = set(sols)

    lodret = solve(f.subs(x, 0).diff(y))
    vandret = solve(f.subs(y, -1).diff(x))
    skrå = solve(f.subs(y, 1 - x).diff(x))

    stat_punkter.update(set([(0, y0) for y0 in lodret]))
    stat_punkter.update(set([(x0, -1) for x0 in vandret]))
    stat_punkter.update(set([(x0, 1 - x0) for x0 in skrå]))
    undersøgelses_punkter = list(stat_punkter) + [(0, 1), (0, -1), (2, -1)]
    f_værdier = [f.subs([(x, x0), (y, y0)]) for x0, y0 in undersøgelses_punkter]

    minimum = min(f_værdier)
    maximum = max(f_værdier)
    assert 3 in f_værdier
    assert S(9) / 4 in f_værdier
    assert S(25) / 8 in f_værdier
    assert S(13) / 4 in f_værdier
    assert 2 in f_værdier
    assert 0 in f_værdier
    u = symbols("u")
    pf = dtuplot.plot3d(
        f, (x, 0, 2), (y, -1, 1), show=False, rendering_kw={"alpha": 0.7}
    )
    # Following command changed, as the scatter function changed. 
    # Each point is now given in a list, so '[]' is added.
    punkter = dtuplot.scatter(
        [Matrix([2, -1, 0]),
        Matrix([1 / 2, 0, 13 / 4])],
        show=False,
        rendering_kw={"color": "red", "s": 20},
    )
    l1 = dtuplot.plot3d_parametric_line(
        u,
        -1,
        f.subs({x: u, y: -1}),
        (u, 0, 2),
        use_cm=False,
        show=False,
        rendering_kw={"color": "red", "linewidth": 2},
    )
    l2 = dtuplot.plot3d_parametric_line(
        0,
        u,
        f.subs({x: 0, y: u}),
        (u, -1, 1),
        use_cm=False,
        show=False,
        rendering_kw={"color": "red", "linewidth": 2},
    )
    l3 = dtuplot.plot3d_parametric_line(
        u,
        1 - u,
        f.subs({x: u, y: 1 - u}),
        (u, 0, 2),
        use_cm=False,
        show=False,
        rendering_kw={"color": "red", "linewidth": 2},
    )
    combined = pf + punkter + l1 + l2 + l3
    combined.camera = {"azim": 118, "elev": 61}
    combined.legend = False


def test_week4():
    u, x, y, a, b = symbols("u x y a b")
    f = exp(u) * sin(u)
    p = dtuplot.plot(f, (u, -pi, pi), {"color": "red"}, axis_center="auto", show=False)
    area = 0
    for i in range(10):
        xval = i * pi / 5 - pi
        yval = f.subs(u, xval)

        # Hjørner af firkant
        c1 = Point(xval, 0)
        c2 = Point(xval + pi / 5, 0)
        c3 = Point(xval + pi / 5, yval)
        c4 = Point(xval, yval)

        poly = Polygon(c1, c2, c3, c4)
        p.extend(
            dtuplot.plot_geometry(poly, {"color": "green", "alpha": 0.5}, show=False)
        )

        # Kan ikke finde areal af en linje
        if yval != 0:
            area += poly.area

    p.legend = False
    assert abs(area.evalf() - 10.799) < 1e-3
    F = integrate(f, u)
    assert F == E**u * sin(u) / 2 - E**u * cos(u) / 2
    assert abs(integrate(f, (u, -pi, pi)).evalf() - 11.5487) < 1e-3
    r = Matrix([sin(u), sin(u) * cos(u)])
    p_kurve = dtuplot.plot_parametric(
        r[0],
        r[1],
        (u, 0, 2 * pi),
        use_cm=False,
        label="r(u)",
        axis_center="auto",
        show=False,
    )
    dr = r.diff(u)
    t = symbols("t")
    r_tan = r.subs(u, pi / 3) + t * dr.subs(u, pi / 3)
    p_point = dtuplot.scatter(r.subs(u, pi / 3), show=False)
    p_tan = dtuplot.plot_parametric(
        r_tan[0], r_tan[1], (t, -1, 1), use_cm=False, label="r '(pi/3)", show=False
    )
    res = p_kurve + p_point + p_tan
    jacobi = sqrt(dr.dot(dr))
    assert simplify(jacobi) == sqrt(cos(u) ** 2 + cos(2 * u) ** 2)
    assert abs(integrate(jacobi, (u, 0, 2 * pi)).evalf() - 6.09722) < 1e-3
    x, y, z = symbols("x y z")
    f = sqrt(x**2 + y**2 + z**2)
    r = Matrix([u * cos(u), u * sin(u), u])
    p_rumkurve = dtuplot.plot3d_parametric_line(
        r[0],
        r[1],
        r[2],
        (u, 0, 2 * pi),
        use_cm=False,
        label="r(u)",
        aspect="equal",
        show=False,
    )
    dr = r.diff(u)
    jacobi = sqrt(dr.dot(dr))
    assert simplify(jacobi) == sqrt(u**2 + 2)
    assert (
        abs(
            integrate(f.subs({x: r[0], y: r[1], z: r[2]}) * jacobi, (u, 0, 5)).evalf()
            - 64.802889
        )
        < 1e-3
    )
    assert abs(integrate(jacobi, (u, 0, 5)).evalf() - 14.9658) < 1e-3


def test_week5():
    # Store dag
    u, v = symbols("u v")
    dtuplot.plot_parametric(
        3 + v, 2 + 2 * v, (v, -1, 1), xlim=(0, 5), ylim=(0, 5), show=False
    )
    x, y = symbols("x y")
    with pytest.warns(
        UserWarning,
        match="The provided expression contains Boolean functions. In order to plot the expression, the algorithm automatically switched to an adaptive sampling.",
    ):
        område = dtuplot.plot_implicit(
            Eq(x, 2),
            Eq(y, 2 * x + 1),
            Eq(y, 1 - x),
            (x <= 2) & (y <= 2 * x + 1) & (y >= 1 - x),
            (x, -0.1, 5),
            (y, -1.1, 5.1),
            aspect="equal",
            size=(8, 8),
            show=False,
        )
    AB = dtuplot.quiver(
        Matrix([1, 0]), Matrix([0, 3]), show=False, rendering_kw={"color": "black"}
    )
    område.extend(AB)
    dtuplot.plot3d_parametric_surface(
        u,
        1 - u + 3 * u * v,
        0,
        (u, 0, 2),
        (v, 0, 1),
        camera={"elev": 90, "azim": -90},
        show=False,
    )
    cirkel = dtuplot.plot_implicit(
        Eq((x - 2) ** 2 + (y - 1) ** 2, 1),
        ((x - 2) ** 2 + (y - 1) ** 2 <= 1),
        (x, 0, 3),
        (y, 0, 3),
        show=False,
    )
    AB = dtuplot.quiver(
        Matrix([2, 1]),
        Matrix([cos(pi / 4), sin(pi / 4)]),
        rendering_kw={"color": "black"},
        show=False,
    )
    cirkel.extend(AB)
    dtuplot.plot3d_parametric_surface(
        2 + v * cos(u),
        1 + v * sin(u),
        0,
        (u, 0, 2 * pi),
        (v, 0, 1),
        camera={"elev": 90, "azim": -90},
        show=False,
    )
    with pytest.warns(
        UserWarning,
        match="The provided expression contains Boolean functions. In order to plot the expression, the algorithm automatically switched to an adaptive sampling.",
    ):
        dtuplot.plot_implicit(
            Eq(y, 1),
            Eq(x, 1),
            Eq(y, x + 1),
            ((x <= 1) & (y >= 1) & (y <= 1 + x)),
            (x, 0, 2),
            (y, 0, 2),
            title="Integrationsområde B",
            show=False,
        )
    r = Matrix([u, 1 + u * v])
    f = 2 * x * y
    JacobiM = Matrix.hstack(r.diff(u), r.diff(v))
    Jacobi = abs(JacobiM.det())
    assert (
        integrate(f.subs([(x, u), (y, 1 + u * v)]) * Jacobi, (u, 0, 1), (v, 0, 1))
        == S(11) / 12
    )
    f = (x - 1) ** 2 * (y + 1) ** 2
    r = Matrix([u * cos(v), u * sin(v) / 2])
    dtuplot.plot3d_parametric_surface(
        u,
        v,
        0,
        (u, 1, 2),
        (v, -pi, pi),
        camera={"elev": 90, "azim": -90},
        xlim=(0, 3),
        ylim=(-3.5, 3.5),
        show=False,
    )
    område = dtuplot.plot3d_parametric_surface(
        r[0],
        r[1],
        0,
        (u, 1, 2),
        (v, -pi, pi),
        camera={"elev": 90, "azim": -90},
        xlim=(-2, 2),
        ylim=(-2, 2),
        show=False,
    )
    JacobiM = Matrix.hstack(r.diff(u), r.diff(v))
    Jacobi = abs(JacobiM.det())
    assert Jacobi.simplify() == abs(u) / 2
    integrand = f.subs([(x, r[0]), (y, r[1])]) * Jacobi
    M = integrate(integrand, (u, 1, 2), (v, -pi, pi))
    assert M == S(267) * pi / 64
    Mmidtpunkt = (
        Matrix(
            [
                integrate(r[0] * integrand, (u, 1, 2), (v, -pi, pi)),
                integrate(r[1] * integrand, (u, 1, 2), (v, -pi, pi)),
            ]
        )
        / M
    )
    assert Mmidtpunkt == Matrix([-S(94) / 89, S(34) / 89])
    punkt = dtuplot.scatter(
        Matrix([Mmidtpunkt[0], Mmidtpunkt[1], 0]),
        show=False,
        rendering_kw={"color": "black"},
        xlim=(-2, 2),
        ylim=(-2, 2),
    )
    område.extend(punkt)
    z = symbols("z")
    f = 8 * z
    r = Matrix([u * cos(u), u * sin(u), u * v])
    dtuplot.plot3d_parametric_surface(*r, (u, 0, pi / 2), (v, 0, 1), show=False)
    kryds = r.diff(u).cross(r.diff(v))
    Jacobi = sqrt((kryds.T * kryds)[0]).simplify()
    assert Jacobi == sqrt(u**4 + u**2)
    integrand = f.subs(z, r[2]) * Jacobi
    assert (
        abs(integrate(integrand, (u, 0, pi / 2), (v, 0, 1)).evalf() - 8.062867) < 1e-3
    )

    # Lille dag
    x, y, z, u, v = symbols("x y z u v", real=True)
    # sympy kan ofte lave bedre simplificeringer af udtryk, når den ved, at variablerne er reelle,
    # så når vi kan er det en god ide at gøre. Ellers kan nogle integraler tage ret lang tid.
    h = x**2 - y**2 + 5
    r = Matrix([x, y, h])
    flade = dtuplot.plot3d_parametric_surface(
        *r,
        (x, -2, 2),
        (y, -2, 2),
        use_cm=True,
        camera={"elev": 30, "azim": -130},
        show=False
    )
    kvadrat = dtuplot.plot3d_parametric_surface(
        x, y, 0, (x, -2, 2), (y, -2, 2), rendering_kw={"color": "black"}, show=False
    )
    res = flade + kvadrat
    r = Matrix([u * cos(v), u * sin(v), h.subs({x: u * cos(v), y: u * sin(v)})])
    flade = dtuplot.plot3d_parametric_surface(
        *r, (u, 1, 2), (v, 0, 2 * pi), show=False, camera={"azim": -75, "elev": 40}
    )
    skive = dtuplot.plot3d_parametric_surface(
        r[0], r[1], 0, (u, 1, 2), (v, 0, 2 * pi), show=False
    )
    res = flade + skive
    dtuplot.plot_parametric(
        u**2 + 1, u, (u, -1, 1), use_cm=False, aspect="equal", show=False
    )
    profil = Matrix([u**2 + 1, 0, u])
    profil_plot = dtuplot.plot3d_parametric_line(
        *profil, (u, -1, 1), show=False, camera={"azim": -65, "elev": 12}, use_cm=False
    )
    r = Matrix([(u**2 + 1) * cos(v), (u**2 + 1) * sin(v), u])
    r = Matrix([[cos(v), -sin(v), 0], [sin(v), cos(v), 0], [0, 0, 1]]) * profil
    flade = dtuplot.plot3d_parametric_surface(
        *r,
        (u, -1, 1),
        (v, 0, 2 * pi),
        camera={"azim": -90, "elev": 15},
        title="Omdrejningsfladen",
        show=False
    )
    dtuplot.plot(
        (x - 1) ** 3 + 1, (x, 0, 2), axis_center="auto", ylabel="z", show=False
    )
    profil = Matrix([u, 0, (u - 1) ** 3 + 1])
    r = Matrix([[cos(v), sin(v), 0], [sin(v), cos(v), 0], [0, 0, 1]]) * profil
    flade = dtuplot.plot3d_parametric_surface(
        *r,
        (u, 0, 2),
        (v, 0, 2 * pi),
        camera={"azim": -65, "elev": 25},
        use_cm=True,
        legend=False,
        show=False
    )
    profil = Matrix([2 + sin(u), 0, u])
    dtuplot.plot3d_parametric_line(
        *profil,
        (u, 0, 8 * pi),
        use_cm=False,
        camera={"azim": -28, "elev": 13},
        show=False
    )
    r = Matrix([(2 + sin(u)) * cos(v), (2 + sin(u)) * sin(v), u])
    flade = dtuplot.plot3d_parametric_surface(
        *r,
        (u, 0, 8 * pi),
        (v, 0, 2 * pi),
        camera={"azim": -62, "elev": 15},
        use_cm=True,
        legend=False,
        show=False
    )
    lede = Matrix([2 * cos(u), sin(u), 0])
    dtuplot.plot3d_parametric_line(
        *lede, (u, 0, 2 * pi), use_cm=False, ylim=(-2, 2), xlim=(-2, 2), show=False
    )
    r = Matrix([lede[0], lede[1], v])
    flade = dtuplot.plot3d_parametric_surface(
        *r,
        (u, 0, 2 * pi),
        (v, 0, 1),
        ylim=(-2, 2),
        xlim=(-2, 2),
        zlim=(0, 2),
        use_cm=True,
        legend=False,
        show=False
    )
    r = Matrix([2 * cos(u), sin(u), v * (2 * cos(u) + sin(u) ** 2)])
    dtuplot.plot3d_parametric_surface(
        *r,
        (u, 0, 2 * pi),
        (v, 0, 1),
        camera={"azim": -120, "elev": 20},
        use_cm=True,
        legend=False,
        show=False
    )
    h = x**2 / 2 + y + 2
    with pytest.warns(
        UserWarning,
        match="The provided expression contains Boolean functions. In order to plot the expression, the algorithm automatically switched to an adaptive sampling.",
    ):
        dtuplot.plot_implicit(
            (x >= 0) & (x <= 2) & (y >= -2 * x) & (y <= 0),
            (x, 0, 2),
            (y, -4, 0),
            aspect="equal",
            show=False,
        )
    trekant = Matrix([u, v * (-2 * u)])
    u_range = (u, 0, 2)
    v_range = (v, 0, 1)
    r = Matrix([trekant[0], trekant[1], h.subs({x: trekant[0], y: trekant[1]})])
    flade = dtuplot.plot3d_parametric_surface(
        *r,
        u_range,
        v_range,
        show=False,
        rendering_kw={"color": "blue"},
        camera={"azim": -155, "elev": 15}
    )
    grundflade = dtuplot.plot3d_parametric_surface(
        r[0], r[1], 0, u_range, v_range, show=False, rendering_kw={"alpha": 0.6}
    )
    res = flade + grundflade
    kryds = r.diff(u).cross(r.diff(v))
    Jacobi = kryds.norm()
    assert Jacobi.simplify() == 2 * sqrt(u**2 + 2) * abs(u)
    assert abs(integrate(Jacobi, u_range, v_range).evalf() - 7.91234) < 1e-3
    profil = Matrix([3 + cos(u), 0, sin(u)])
    u_range = (u, 0, 2 * pi)
    h = z**2
    dtuplot.plot_parametric(
        profil[0],
        profil[2],
        u_range,
        xlim=(0, 4),
        ylim=(-1, 1),
        ylabel="z",
        use_cm=False,
        aspect="equal",
        show=False,
    )
    r = Matrix([[cos(v), -sin(v), 0], [sin(v), cos(v), 0], [0, 0, 1]]) * profil
    v_range = (v, 0, 2 * pi)
    flade = dtuplot.plot3d_parametric_surface(
        *r, u_range, v_range, zlim=(-4, 4), use_cm=True, legend=False, show=False
    )
    Jacobi = r.diff(u).cross(r.diff(v)).norm().simplify()
    assert Jacobi == abs(cos(u) + 3)
    integrand = h.subs({x: r[0], y: r[1], z: r[2]}) * Jacobi
    assert integrate(integrand, u_range, v_range) == 6 * pi**2
    integrand = Jacobi
    assert integrate(integrand, u_range, v_range) == 12 * pi**2


def test_week6():
    ## Tænk ikke for meget over dette

    x, y, z, u, v, w = symbols("x y z u v w", real=True)
    ur = (u, 0, 1)
    l1 = dtuplot.plot3d_parametric_line(
        u, 0, 0, ur, show=False, use_cm=False, rendering_kw={"color": "red"}
    )
    l2 = dtuplot.plot3d_parametric_line(
        0, u, 0, ur, show=False, use_cm=False, rendering_kw={"color": "red"}
    )
    l3 = dtuplot.plot3d_parametric_line(
        0, 0, u, ur, show=False, use_cm=False, rendering_kw={"color": "red"}
    )
    l4 = dtuplot.plot3d_parametric_line(
        1 - u, u, 0, ur, show=False, use_cm=False, rendering_kw={"color": "red"}
    )
    l5 = dtuplot.plot3d_parametric_line(
        u, 0, 1 - u, ur, show=False, use_cm=False, rendering_kw={"color": "red"}
    )
    l6 = dtuplot.plot3d_parametric_line(
        0, u, 1 - u, ur, show=False, use_cm=False, rendering_kw={"color": "red"}
    )
    grundflade = dtuplot.plot3d_parametric_surface(
        u * (1 - v),
        v,
        0,
        ur,
        (v, 0, 1),
        show=False,
        rendering_kw={"color": "blue", "alpha": 0.75},
    )
    dash1 = dtuplot.plot3d_parametric_line(
        0.3,
        0.7 * u,
        0,
        ur,
        show=False,
        use_cm=False,
        rendering_kw={"color": "grey", "linestyle": "dashed"},
    )
    dash2 = dtuplot.plot3d_parametric_line(
        0.3,
        0.7 - 0.7 * u,
        0.7 * u,
        ur,
        show=False,
        use_cm=False,
        rendering_kw={"color": "grey", "linestyle": "dashed"},
    )
    dash3 = dtuplot.plot3d_parametric_line(
        0.3,
        0,
        0.7 * u,
        ur,
        show=False,
        use_cm=False,
        rendering_kw={"color": "grey", "linestyle": "dashed"},
    )
    pil3d = dtuplot.quiver(
        Matrix([0.3, 0.375, 0]),
        Matrix([0, 0, 0.375]),
        show=False,
        rendering_kw={"color": "black"},
    )

    with pytest.warns(
        UserWarning,
        match="The provided expression contains Boolean functions. In order to plot the expression, the algorithm automatically switched to an adaptive sampling.",
    ):
        bund = dtuplot.plot_implicit(
            (x >= 0) & (x <= 1) & (y >= 0) & (y <= 1 - x),
            (x, 0, 1),
            (y, 0, 1),
            show=False,
        )
    pil2d = dtuplot.quiver(
        Matrix([0.2, 0]),
        Matrix([0, 0.8]),
        show=False,
        rendering_kw={"color": "black", "width": 0.025},
    )

    skelet = l1 + l2 + l3 + l4 + l5 + l6 + dash1 + dash2 + dash3 + pil3d
    tetra = grundflade + skelet
    tetra.legend = False
    tetra.camera = {"azim": 25, "elev": 25}
    tetra.size = (6, 6)
    trekant = bund + pil2d
    trekant.legend = False
    trekant.grid = False

    d = dtuplot.plotgrid(tetra, trekant, nr=1, nc=2, size=(6, 6), show=False)
    f = (1 - x) ** 2
    h = 1 - x - y
    bund_punkt = Matrix([u, v * (1 - u), 0])
    top_punkt = Matrix([u, v * (1 - u), h.subs({x: u, y: v * (1 - u)})])
    r = bund_punkt + w * (top_punkt - bund_punkt)
    _ = Matrix.hstack(*[r.diff(v) for v in [u, v, w]])
    determinanten = _.det()  # underscore holder værdien fra seneste output
    Jacobi = determinanten
    # vi ved at determinanten er positiv i det betragtede område, så det er ikke nødvendigt at tage absolut værdien
    # Jacobi.simplify()
    assert Jacobi.factor() == -((u - 1) ** 2 * (v - 1)).factor()
    M = integrate(
        f.subs({x: r[0], y: r[1], z: r[2]}) * Jacobi, (u, 0, 1), (v, 0, 1), (w, 0, 1)
    )
    assert M == S(1) / 10
    xi = (
        Matrix(
            [
                integrate(
                    coord * f.subs({x: r[0], y: r[1], z: r[2]}) * Jacobi,
                    (u, 0, 1),
                    (v, 0, 1),
                    (w, 0, 1),
                )
                for coord in r
            ]
        )
        / M
    )
    assert xi == Matrix([S(1) / 6, S(5) / 18, S(5) / 18])
    with pytest.warns(
        UserWarning,
        match="The provided expression contains Boolean functions. In order to plot the expression, the algorithm automatically switched to an adaptive sampling.",
    ):
        dtuplot.plot_implicit(
            (x >= 1 / 2) & (x <= 2) & (y <= 1 / x) & (y >= 0),
            (x, 0, 2),
            (y, 0, 2),
            show=False,
        )
    profil = Matrix([u, 0, v / u])
    legeme = Matrix([[cos(w), -sin(w), 0], [sin(w), cos(w), 0], [0, 0, 1]]) * profil
    determinanten = Matrix.hstack(*[legeme.diff(v) for v in [u, v, w]]).det()
    Jacobi = abs(determinanten.simplify())
    assert Jacobi == 1
    assert integrate(Jacobi, (u, S(1) / 2, 2), (v, 0, 1), (w, 0, 2 * pi)) == 3 * pi
    l1 = Eq(y, cos(x))
    l2 = Eq(y, cos(x) + 1)
    x_range = (x, 0, 2 * pi)

    dtuplot.plot_implicit(l1, l2, x_range, (y, -1, 2), aspect="equal", show=False)
    mur = Matrix([u, cos(u) + v, w * u / 2])
    determinanten = Matrix.hstack(*[mur.diff(v) for v in [u, v, w]]).det()
    ## antag at variablerne er positive
    Jacobi = abs(posify(determinanten)[0])
    assert determinanten == u / 2
    f = 1 / x
    integrand = (f.subs({x: mur[0], y: mur[1], z: mur[2]}) * Jacobi).simplify()
    integrand = S(1) / 2
    assert integrate(integrand, (u, 0, 2 * pi), (v, 0, 1), (w, 0, 1)) == pi


def test_week7():
    # Store dag
    assert True
    x, y, z, u = symbols("x y z u", real=True)
    r1 = Matrix([cos(u), sin(u), u / 2])
    r2 = Matrix([1, 0, u / 2])
    V = Matrix([-y, x, 2 * z])
    u_range = (u, 0, 4 * pi)

    K1 = dtuplot.plot3d_parametric_line(
        *r1, u_range, show=False, rendering_kw={"color": "red"}
    )
    K2 = dtuplot.plot3d_parametric_line(
        *r2, u_range, show=False, rendering_kw={"color": "blue"}
    )
    felt = dtuplot.plot_vector(
        V,
        (x, -1, 1),
        (y, -1, 1),
        (z, 0, 6),
        show=False,
        quiver_kw={"alpha": 0.5, "length": 0.1, "color": "black"},
        n=5,
    )

    combined = K1 + K2 + felt
    combined.legend = False
    r1d = r1.diff(u)
    r2d = r2.diff(u)
    integrand1 = V.subs({x: r1[0], y: r1[1], z: r1[2]}).dot(r1d)
    integrand2 = V.subs({x: r2[0], y: r2[1], z: r2[2]}).dot(r2d)
    integrand1.simplify(), integrand2.simplify()
    assert integrate(integrand1, (u, 0, 4 * pi)) == 4 * pi + 4 * pi**2
    assert integrate(integrand2, (u, 0, 4 * pi)) == 4 * pi**2

    u_range = (u, 0, 1)
    r1 = dtuplot.plot3d_parametric_line(
        u, 0, 0, u_range, show=False, rendering_kw={"color": "red"}
    )
    r2 = dtuplot.plot3d_parametric_line(
        1, u, 0, u_range, show=False, rendering_kw={"color": "red"}
    )
    r3 = dtuplot.plot3d_parametric_line(
        1, 1, u, u_range, show=False, rendering_kw={"color": "red"}
    )
    xyz = dtuplot.scatter(
        Matrix([1, 1, 1]), show=False, rendering_kw={"color": "black"}
    )
    combined = r1 + r2 + r3 + xyz
    combined.legend = False
    combined.camera = {"azim": 37, "elev": 16}

    r1 = Matrix([u, 0, 0])
    r2 = Matrix([x, u, 0])
    r3 = Matrix([x, y, u])
    r1d = r1.diff(u)
    r2d = r2.diff(u)
    r3d = r3.diff(u)
    r1d, r2d, r3d

    V = Matrix([y**2 + z, 2 * y * z**2 + 2 * y * x, 2 * y**2 * z + x])
    integrand1 = V[0].subs({x: u, y: 0, z: 0})
    integrand2 = V[1].subs({y: u, z: 0})
    integrand3 = V[2].subs({z: u})
    assert integrand1 == 0
    assert integrand2 == 2 * u * x
    assert integrand3 == 2 * u * y**2 + x
    F = (
        integrate(integrand1, (u, 0, x))
        + integrate(integrand2, (u, 0, y))
        + integrate(integrand3, (u, 0, z))
    )
    V_F = dtutools.gradient(F)
    assert Eq(V_F, V) == True
    t = symbols("t")
    knude = (
        Matrix(
            [
                -10 * cos(t) - 2 * cos(5 * t) + 15 * sin(2 * t),
                -15 * cos(2 * t) + 10 * sin(t) - 2 * sin(5 * t),
                10 * cos(3 * t),
            ]
        )
        * S(1)
        / 10
    )
    dtuplot.plot3d_parametric_line(
        *knude, (t, -pi, pi), rendering_kw={"color": "blue"}, legend=False, show=False
    )
    assert (
        integrate(
            V.subs({x: knude[0], y: knude[1], z: knude[2]}).dot(knude.diff(t)),
            (t, -pi, pi),
        )
        == 0
    )

    # Lille dag
    x, y, z = symbols("x y z")
    f = x**2 * sin(y) * z
    df = dtutools.gradient(f)
    dtutools.gradient(f, [x, z])
    df0 = df.subs([(x, 1), (y, 2), (z, 3)])
    assert abs(N(df0)[0] - 5.45578) < 1e-4
    assert abs(N(df0)[1] + 1.2484) < 1e-4
    assert abs(N(df0)[2] - 0.909297) < 1e-4
    V = Matrix([x * y * sin(z), x * ln(y), -x * y * z**3])
    divV = dtutools.div(V, var=[x, y, z])
    assert divV == -3 * x * y * z**2 + x / y + y * sin(z)
    u, v, w = symbols("u v w")
    # v optræder ikke i vektorfeltet
    V2 = Matrix([u, w**2, u * sin(w)])
    assert dtutools.div(V2, var=[u, v, w]) == u * cos(w) + 1
    divV0 = divV.subs([(x, 1), (y, 2), (z, 3)])
    N(divV0)
    rotV = dtutools.rot(V, var=[x, y, z])
    assert rotV == Matrix(
        [-x * z**3, x * y * cos(z) + y * z**3, -x * sin(z) + log(y)]
    )
    assert dtutools.rot(V2, var=[u, v, w]) == Matrix([-2 * w, -sin(w), 0])
    rotV0 = rotV.subs([(x, 1), (y, 2), (z, 3)])
    N(rotV0)


def test_week8():
    x, y, z = symbols("x y z")
    V = Matrix([x, y, 2])
    u, v = symbols("u v")
    r = Matrix([u * cos(v), u * sin(v), u**2])
    p_parab = dtuplot.plot3d_parametric_surface(
        *r,
        (u, 0, 1),
        (v, -pi, pi),
        use_cm=False,
        camera={"elev": 25, "azim": -55},
        show=False
    )
    p_felt = dtuplot.plot_vector(
        V,
        (x, -1, 1),
        (y, -1, 1),
        (z, 0, 1.2),
        n=4,
        use_cm=False,
        quiver_kw={"alpha": 0.5, "length": 0.1, "color": "red"},
        show=False,
    )
    res = p_parab + p_felt
    # Kommer vektorer fra alle hjørner. Der kommer for mange
    # vektorer hvis "p_parab" bruges, så her sættes n ned
    dummy_surface = dtuplot.plot3d_parametric_surface(
        *r, (u, 0, 1), (v, -pi, pi), n=8, show=False
    )
    p_felt2 = dtuplot.plot_vector(
        V,
        (x, -1, 1),
        (y, -1, 1),
        (z, 0, 1),
        slice=dummy_surface[0],
        use_cm=False,
        quiver_kw={"alpha": 0.5, "length": 0.1, "color": "red"},
        show=False,
    )
    # bedre kamera vinkel til at se vektorerne her
    p_parab.camera = {"elev": 60, "azim": -50}
    res = p_parab + p_felt2
    N_parab = r.diff(u).cross(r.diff(v))
    simplify(N_parab)
    N_parab = -simplify(N_parab)
    N_parab.norm().simplify()
    integrand = N_parab.dot(V.subs({x: r[0], y: r[1], z: r[2]}))
    assert simplify(integrand) == 2 * u * (u**2 - 1)
    assert integrate(integrand, (u, 0, 1), (v, -pi, pi)) == -pi
    r2 = Matrix([u * cos(v), u * sin(v), 1])
    n2 = r2.diff(u).cross(r2.diff(v))
    assert simplify(n2) == Matrix([0, 0, u])
    integrand2 = n2.dot(V.subs({x: r2[0], y: r2[1], z: r2[2]}))
    assert integrate(integrand2, (u, 0, 1), (v, -pi, pi)) == 2 * pi
    p_låg = dtuplot.plot3d_parametric_surface(
        *r2,
        (u, 0, 1),
        (v, -pi, pi),
        use_cm=False,
        n1=4,
        n2=16,
        camera={"elev": 25, "azim": -55},
        show=False
    )
    p_feltlåg = dtuplot.plot_vector(
        V,
        (x, -1, 1),
        (y, -1, 1),
        (z, 0, 1),
        slice=p_låg[0],
        use_cm=False,
        quiver_kw={"alpha": 0.5, "length": 0.1, "color": "red"},
        show=False,
    )
    p_parab.camera = {"elev": 15, "azim": -60}
    res = p_parab + p_felt2 + p_låg + p_feltlåg
    V = Matrix([-x + cos(z), -y * x, 3 * z + exp(y)])
    a = symbols("a")
    # Da man ikke kan plotte et 3D volumen (nemt), plotter vi overfladerne,
    p = dtuplot.plot3d_parametric_surface(
        x,
        y,
        y**2,
        (x, 0, 3),
        (y, 0, 2),
        {"color": "royalblue"},
        use_cm=False,
        aspect="equal",
        show=False,
    )
    p.extend(
        dtuplot.plot3d_parametric_surface(
            0,
            y,
            a * y**2,
            (a, 0, 1),
            (y, 0, 2),
            {"color": "royalblue", "alpha": 0.5},
            use_cm=False,
            aspect="equal",
            show=False,
        )
    )
    p.extend(
        dtuplot.plot3d_parametric_surface(
            3,
            y,
            a * y**2,
            (a, 0, 1),
            (y, 0, 2),
            {"color": "royalblue", "alpha": 0.5},
            use_cm=False,
            aspect="equal",
            show=False,
        )
    )
    p.extend(
        dtuplot.plot3d_parametric_surface(
            x,
            2,
            a * 4,
            (x, 0, 3),
            (a, 0, 1),
            {"color": "royalblue", "alpha": 0.5},
            use_cm=False,
            aspect="equal",
            show=False,
        )
    )
    p.extend(
        dtuplot.plot3d_parametric_surface(
            x,
            y,
            0,
            (x, 0, 3),
            (y, 0, 2),
            {"color": "royalblue", "alpha": 0.5},
            use_cm=False,
            aspect="equal",
            show=False,
        )
    )
    p_felt = dtuplot.plot_vector(
        V,
        (x, 0, 3),
        (y, 0, 2),
        (z, 0, 4),
        n=4,
        use_cm=False,
        quiver_kw={"alpha": 0.5, "length": 0.05, "color": "red"},
        show=False,
    )
    res = p + p_felt
    u, v, w = symbols("u v w")
    r = Matrix([u, v, w * v**2])
    M = Matrix.hstack(r.diff(u), r.diff(v), r.diff(w))
    jacobi = M.det()
    assert jacobi == v**2
    divV = dtutools.div(V, var=[x, y, z])
    assert divV == 2 - x
    divV_r = divV.subs({x: r[0], y: r[1], z: r[2]})
    assert divV_r == 2 - u
    assert integrate(divV_r * jacobi, (u, 0, 3), (v, 0, 2), (w, 0, 1)) == 4
    x, y, z = symbols("x y z")
    V = Matrix([-y + x, x, z])
    u, v = symbols("u v")
    radius = 2
    r = radius * Matrix([sin(u) * cos(v), sin(u) * sin(v), cos(u)])
    p_surface = dtuplot.plot3d_spherical(
        radius,
        (u, pi / 6, pi / 2),
        (v, 0, pi),
        aspect="equal",
        camera={"elev": 25, "azim": 55},
        show=False,
    )
    dummy_surface = dtuplot.plot3d_spherical(
        radius, (u, pi / 6, pi / 2), (v, 0, pi), show=False, n=8
    )
    p_felt = dtuplot.plot_vector(
        V,
        (x, -1, 1),
        (y, -1, 1),
        (z, 0, 1),
        slice=dummy_surface[0],
        use_cm=False,
        quiver_kw={"alpha": 0.5, "length": 0.2, "color": "red"},
        show=False,
    )
    res = p_surface + p_felt
    N = r.diff(u).cross(r.diff(v))
    integrand = N.dot(V.subs({x: r[0], y: r[1], z: r[2]}))
    assert (
        integrate(integrand, (u, pi / 6, pi / 2), (v, 0, pi)) == S(5) * sqrt(3) * pi / 2
    )
    V = Matrix([8 * x, 8, 4 * z**3])
    r = u * Matrix([sin(v) * cos(w), sin(v) * sin(w), cos(v)])
    M = Matrix.hstack(r.diff(u), r.diff(v), r.diff(w))
    # determinanten er positiv. Eneste led der ikke er
    # kvadreret er sin(v), som er positiv indenfor v's grænser
    jacobi = M.det()
    divV = dtutools.div(V, [x, y, z])
    divV_r = divV.subs({x: r[0], y: r[1], z: r[2]})
    assert (
        integrate(divV_r * jacobi, (u, 0, a), (v, 0, pi / 2), (w, -pi, pi))
        == S(8) * pi * a**5 / 5 + S(16) * pi * a**3 / 3
    )
    # Første flade tilsvarer at fastsætte u til radius
    r1 = r.subs(u, a)
    n1 = r1.diff(v).cross(r1.diff(w))  # har tjekket at den pejer udad
    integrand1 = n1.dot(V.subs({x: r1[0], y: r1[1], z: r1[2]}))
    flux1 = integrate(integrand1, (v, 0, pi / 2), (w, -pi, pi))
    # Anden flade tilsvarer at fastsætte v til pi/2
    # Tilbageværende bliver skiven med radius a
    r2 = r.subs(v, pi / 2)
    n2 = -r2.diff(u).cross(r2.diff(w))  # skal være negativ før at den pejer udad
    integrand2 = n2.dot(V.subs({x: r2[0], y: r2[1], z: r2[2]}))
    flux2 = integrate(integrand2, (u, 0, a), (w, -pi, pi))
    assert flux1 + flux2 == S(8) * pi * a**5 / 5 + S(16) * pi * a**3 / 3


def test_week9():
    x, y, z = symbols("x y z")
    V = Matrix([3 * z, 5 * x, -2 * y])
    u = symbols("u")
    s = Matrix([cos(u), sin(u), sin(u)])
    p_kurve = dtuplot.plot3d_parametric_line(
        *s,
        (u, 0, 2 * pi),
        {"color": "orange", "linewidth": 5},
        zlim=(-1.1, 1.1),
        use_cm=False,
        aspect="equal",
        show=False
    )
    p_felt = dtuplot.plot_vector(
        V,
        (x, -1, 1),
        (y, -1, 1),
        (z, -1, 1),
        n=4,
        use_cm=False,
        quiver_kw={"alpha": 0.5, "length": 0.05, "color": "red"},
        show=False,
    )
    res = p_kurve + p_felt
    ds = s.diff(u)
    integrand = ds.dot(V.subs({x: s[0], y: s[1], z: s[2]}))
    assert integrate(integrand, (u, 0, 2 * pi)) == 2 * pi
    u, v = symbols("u v")
    r = Matrix([u * cos(v), u * sin(v), u * sin(v)])
    assert dtutools.rot(V, var=[x, y, z]) == Matrix([-2, 3, 5])
    p_flade = dtuplot.plot3d_parametric_surface(
        *r, (u, 0, 1), (v, 0, 2 * pi), use_cm=False, show=False
    )
    p_felt_rot = dtuplot.plot_vector(
        dtutools.rot(V, var=[x, y, z]),
        (x, -1, 1),
        (y, -1, 1),
        (z, -1, 1),
        n=4,
        use_cm=False,
        quiver_kw={"alpha": 0.5, "length": 0.05, "color": "red"},
        show=False,
    )
    res = p_kurve + p_flade + p_felt_rot
    N = r.diff(u).cross(r.diff(v))
    assert simplify(N) == Matrix([0, -u, u])
    p_normalvektor = dtuplot.quiver(s.subs(u, 0), N.subs({u: 1, v: 0}), show=False)
    p_omloebs = dtuplot.quiver(s.subs(u, 0), ds.subs(u, 0), show=False)
    p_kurve.camera = {"elev": 5, "azim": -35}
    res = p_kurve + p_flade + p_normalvektor + p_omloebs
    integrand = N.dot(dtutools.rot(V, var=[x, y, z]))
    assert simplify(integrand) == 2 * u
    assert integrate(integrand, (v, 0, 2 * pi), (u, 0, 1)) == 2 * pi
    x, y, z, u, v = symbols("x y z u v")
    V = Matrix([y * x, y * z, x * z])
    r = Matrix([cos(v) * (1 + u**2), sin(v) * (1 + u**2), sin(u)])
    p_flade = dtuplot.plot3d_parametric_surface(
        *r,
        (u, -pi / 2, pi / 2),
        (v, -pi, pi / 2),
        use_cm=False,
        aspect="equal",
        camera={"elev": 30, "azim": -80},
        show=False
    )
    dummy_surface = dtuplot.plot3d_parametric_surface(
        *r, (u, -pi / 2, pi / 2), (v, -pi, pi / 2), n=10, show=False
    )
    p_felt_flade = dtuplot.plot_vector(
        V,
        (x, -1, 1),
        (y, -1, 1),
        (z, -1, 1),
        slice=dummy_surface[0],
        quiver_kw={"alpha": 0.5, "length": 0.2, "color": "red"},
        use_cm=False,
        show=False,
    )
    # Vi vil gerne kende retningen af normalvektoren, så vi
    # sikrer os at vi gennemløber randen i den rigtige retning
    N = r.diff(u).cross(r.diff(v))
    # vektor dobbelt så lang, så den kan ses fra denne vinkel
    p_normalvektor = dtuplot.quiver(
        r.subs({u: 0, v: -pi / 4}),
        2 * N.subs({u: 0, v: -pi / 4}),
        {"color": "black"},
        show=False,
    )
    combined_flade = p_flade + p_felt_flade + p_normalvektor
    combined_flade.title = "Fladen og feltet"
    s1 = r.subs(u, -pi / 2)
    s2 = r.subs(u, pi / 2)
    s3 = r.subs(v, -pi)
    s4 = r.subs(v, pi / 2)
    ds1 = s1.diff(v)
    ds2 = s2.diff(v)
    ds3 = s3.diff(u)
    ds4 = s4.diff(u)
    p_rand = dtuplot.plot3d_parametric_line(
        *s1,
        (v, -pi, pi / 2),
        {"color": "orange", "linewidth": 5},
        use_cm=False,
        aspect="equal",
        camera={"elev": 30, "azim": -80},
        show=False
    )
    p_rand.extend(
        dtuplot.plot3d_parametric_line(
            *s2,
            (v, -pi, pi / 2),
            {"color": "orange", "linewidth": 5},
            use_cm=False,
            show=False
        )
    )
    p_rand.extend(
        dtuplot.plot3d_parametric_line(
            *s3,
            (u, -pi / 2, pi / 2),
            {"color": "orange", "linewidth": 5},
            use_cm=False,
            show=False
        )
    )
    p_rand.extend(
        dtuplot.plot3d_parametric_line(
            *s4,
            (u, -pi / 2, pi / 2),
            {"color": "orange", "linewidth": 5},
            use_cm=False,
            show=False
        )
    )
    p_flade_mesh = dtuplot.plot3d_parametric_surface(
        *r,
        (u, -pi / 2, pi / 2),
        (v, -pi, pi / 2),
        rendering_kw={"alpha": 0},
        wf_rendering_kw={"alpha": 0.4},
        wireframe=True,
        show=False
    )
    p_felt = dtuplot.plot_vector(
        V,
        (x, -3, 3),
        (y, -3, 3),
        (z, -1, 1),
        quiver_kw={"alpha": 0.5, "length": 0.2, "color": "red"},
        n1=5,
        n2=5,
        n3=2,
        show=False,
    )
    # for hver af de 4 felter vil vi også gerne have retningen af deres afledte
    # dette skal bruges til at bestemme hvilken retning vi skal omløbe linjerne i
    p_omloebs = dtuplot.quiver(
        s1.subs({u: -pi / 2, v: -pi}),
        ds1.subs({u: -pi / 2, v: -pi}),
        {"color": "red"},
        show=False,
    )
    p_omloebs.extend(
        dtuplot.quiver(
            s2.subs({u: -pi / 2, v: -pi}),
            ds2.subs({u: -pi / 2, v: -pi}),
            {"color": "green"},
            show=False,
        )
    )
    p_omloebs.extend(
        dtuplot.quiver(
            s3.subs({u: -pi / 2, v: -pi}),
            ds3.subs({u: -pi / 2, v: -pi}),
            {"color": "blue"},
            show=False,
        )
    )
    p_omloebs.extend(
        dtuplot.quiver(
            s4.subs({u: -pi / 2, v: -pi}),
            ds4.subs({u: -pi / 2, v: -pi}),
            {"color": "purple"},
            show=False,
        )
    )
    combined_rand = p_rand + p_flade_mesh + p_felt + p_omloebs
    combined_rand.legend = False
    combined_rand.title = "Feltet på randen af fladen"
    dtuplot.plotgrid(combined_rand, combined_flade, nr=1, nc=2, show=False)
    i1 = ds1.dot(V.subs({x: s1[0], y: s1[1], z: s1[2]}))
    i2 = ds2.dot(V.subs({x: s2[0], y: s2[1], z: s2[2]}))
    i3 = ds3.dot(V.subs({x: s3[0], y: s3[1], z: s3[2]}))
    i4 = ds4.dot(V.subs({x: s4[0], y: s4[1], z: s4[2]}))
    assert (
        simplify(
            -integrate(i1, (v, -pi, pi / 2))
            + integrate(i2, (v, -pi, pi / 2))
            + integrate(i3, (u, -pi / 2, pi / 2))
            - integrate(i4, (u, -pi / 2, pi / 2))
        )
        == -S(5) * pi**2 / 2 + pi**4 / 16 + 21
    )
    integrand = N.dot(dtutools.rot(V, var=[x, y, z]).subs({x: r[0], y: r[1], z: r[2]}))
    simplify(integrand)
    assert (
        integrate(integrand, (u, -pi / 2, pi / 2), (v, -pi, pi / 2))
        == -S(5) * pi**2 / 2 + pi**4 / 16 + 21
    )



from spb import MB, PB, BB, KB, MAB
from spb.defaults import THREE_D_B, TWO_D_B
from spb.functions import _set_labels
from spb.series import VectorBase
from spb.utils import _instantiate_backend
from sympy import Matrix, latex, symbols
from sympy.external import import_module
from spb.backends.base_backend import Plot
import warnings

np = import_module("numpy")


class ArrowSeries(VectorBase):
    """Represent a vector field."""

    is_vector = True
    is_slice = False
    is_streamlines = False
    _allowed_keys = []

    def __init__(self, start, direction, label=None, **kwargs):
        # Ranges must be given for VectorBase, even though they are None
        super().__init__(
            [start, direction], ranges=start.shape[-1] * [None], label=label, **kwargs
        )

        self.start = start
        self.direction = direction

        self._label = f"{start}->{direction}" if label is None else label
        self._latex_label = latex(f"{start}->{direction}") if label is None else label

        # Standard for 'use_cm' should be False
        self.use_cm = kwargs.get("use_cm", False)
        # Linked colormap using vector2d renderer
        self.use_quiver_solid_color = not self.use_cm
        # Line color needed for Mayavi
        self._line_color = kwargs.get("line_color", None)
        # _sal argument saved here and passed to matplotlib backend (if used)
        self._sal = kwargs.get("_sal", False)

    def __str__(self):
        # Overwrite the VectorBase __str__ as it assumes things
        # about variables that does not hold for this class.
        return self._str_helper(
            f"Arrow Series with start point {self.start}, and direction {self.direction}"
        )

    def get_data(self):
        # This format works for both MB and PB
        # Has to translate to start/end and transpose
        # such that the x,y,z lims match the arrow
        # compensation for this in arrow done in
        # quiverplot_helpers.
        start = np.array(self.start)
        end = start + np.array(self.direction)
        return np.array(
            [start, end]
        ).T  # [np.array([v]) for v in list(self.start) + list(self.direction)]


# Specify 2D class such that this can be linked with renderer
class Arrow2DSeries(ArrowSeries):
    is_2Dvector = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


# Specify 3D class such that this can be linked with renderer
class Arrow3DSeries(ArrowSeries):
    is_3Dvector = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def quiver(*args, **kwargs):
    """Create a plot with a vector.

    Args:
        start (MatrixBase, np.ndarray, list, float): The starting coordinates (2D or 3D) of the vector. Can be given in multitude of ways/inputs.
        direction (MatrixBase, np.ndarray, list, float): The direction (2D or 3D) of the vector. Can be given in multitude of ways/inputs.
        rendering_kw (dict, optional): A dictionary forwarded to dtuplot.plot(), see SPB docs for reference.
        qlim (bool, optional): Boolean relevant only for backend MB. If 'True' (default) adjusts xlim and ylim to the quiver.
        color (str, optional): A string to set the color of the vector with. With no argument color = 'blue'.
        show (bool, optional): Boolean, if 'True': show plot, other just return object without plotting. Defaults to 'True'.

    Returns:
        Plot: The vector plot.
    """
    # format if numbers are entered directly instead of lists
    num_single = 0
    args = list(args)
    for i in range(len(args)):
        if type(args[i]) in [int, float]:
            num_single += 1
        else:
            try:
                args[i] = float(args[i])
                num_single += 1
            except:
                break
    if num_single == 4:
        if len(args) > 4:
            args = [[args[0], args[1]], [args[2], args[3]], *args[4:]]
        else:
            args = [[args[0], args[1]], [args[2], args[3]]]
    elif num_single == 6:
        if len(args) > 6:
            args = [[args[0], args[1], args[2]], [args[3], args[4], args[5]], *args[6:]]
        else:
            args = [[args[0], args[1], args[2]], [args[3], args[4], args[5]]]
    elif num_single != 0:
        raise ValueError(
            f"Error! Wrong format used in quiver. Got {num_single} arguments that could be start or direction vector coordinates!"
        )

    # split arguments into vectors and other arguments
    point_args = []
    otherargs = []
    for i in range(len(args)):
        if type(args[i]) in [list, np.ndarray, tuple, type(Matrix())]:
            newpoint = np.array(args[i]).flatten()
            point_args.append(newpoint)
        elif type(args[i]) == dict:
            kwargs.setdefault("rendering_kw", args[i])
        else:
            otherargs.append(args[i])
    assert (
        len(point_args) == 2
    ), f"Error! Start and direction vectors must be provided (only two vectors)!"

    try:
        point_args = np.array(point_args, dtype=float)
        assert point_args.shape[-1] in [
            2,
            3,
        ], "Error! Start and direction vectors must be 2D or 3D!"
        args = otherargs
    except:
        raise ValueError(
            f"Error! Wrong format used in quiver. Got {point_args[0]} as starting point(s) and {point_args[1]} as ending point(s)!"
        )

    # want structure to be [list of starts, list of ends]
    # where list of starts could be [start1, start2, ...], either 2D or 3D points
    if len(point_args.shape) == 2:
        point_args = point_args[:, None, :]

    labels = kwargs.pop("label", [])
    if type(labels) == str:
        labels = [labels]
    assert type(labels) == list, f"Error! Label must be a list or a string!"
    assert len(labels) in [
        0,
        point_args.shape[1],
    ], f"Error! Number of labels must be equal to number of arrows, or empty list!"
    if labels == []:
        labels = [None] * point_args.shape[1]

    kwargs.setdefault("legend", True)
    # params are directly linked with interactive plots
    # if present, the BaseSeries will activate the 'is_interactive' flag
    params = kwargs.get("params", None)
    is_interactive = False if params is None else True
    if is_interactive:
        raise NotImplementedError("Interactive quiver plots are not yet implemented!")
        from spb.interactive import iplot

        kwargs["is_vector"] = True
        return iplot(*args, **kwargs)

    # rendering_kw needs to be passed to the plotting backend, but not
    # to the series. Thus pulled out here.
    # Create if it does not exist
    rendering_kw = kwargs.pop("rendering_kw", {})
    # normalize argument needs to be passed to series object, but not to
    # backend. Otherwise warning will be raised. Thus pulled out here.
    normalize = kwargs.pop("normalize", False)
    # automatically adjust the limits to the quiver. Only relevant for MB.
    # saved in series, which propagates to the renderer when created.
    qlim = kwargs.pop("qlim", True)

    # if 2D
    if point_args.shape[-1] == 2:
        Backend = kwargs.pop("backend", TWO_D_B)
        Series = Arrow2DSeries

        # Specific for matplotlib backend
        if Backend == MB:
            # Update values for length of vector to fit
            rendering_kw.setdefault("angles", "xy")
            rendering_kw.setdefault("scale_units", "xy")
            rendering_kw.setdefault("scale", 1)
        elif Backend == PB:
            rendering_kw.setdefault("scale", 1)
            rendering_kw.setdefault("scaleratio", 1)
        elif Backend == BB:
            mag = np.linalg.norm(point_args[-1].flatten(), 2)
            rendering_kw.setdefault("scale", mag)
            rendering_kw.setdefault("pivot", "tail")
        elif Backend == KB:
            raise NotImplementedError("K3D backend does not support 2D vector plots!")
        elif Backend == MAB:
            raise NotImplementedError(
                "Mayavi backend does not support 2D vector plots!"
            )
    else:
        Backend = kwargs.pop("backend", THREE_D_B)
        Series = Arrow3DSeries

        if Backend == PB:
            rendering_kw.setdefault("sizeref", 1)
            rendering_kw.setdefault("sizemode", "scaled")
            rendering_kw.setdefault("anchor", "tail")  # point_args[0,0]
        elif Backend == BB:
            raise NotImplementedError("Bokeh backend does not support 3D vector plots!")
        elif Backend == KB:
            rendering_kw.setdefault("pivot", "tail")
        elif Backend == MAB:
            warnings.warn(f"Mayavi is not currently supported by dtumathtools.")
            rendering_kw.setdefault("scale_factor", 1)
            rendering_kw.setdefault("resolution", 100)
            display_warning = kwargs.pop("warning", True)
            if display_warning:
                warnings.warn(
                    f"Because of the Mayavi backend, the origin of the vector might be slightly off. To supress this warning, set 'warning=False'"
                )

    series = [
        Series(
            start,
            stop,
            *otherargs,
            label=label,
            normalize=normalize,
            _sal=(qlim and Backend==MB), # save qlim kwarg in series object (if MB)
            **kwargs,
        ) for start, stop, label in zip(point_args[0, :, :], point_args[1, :, :], labels)
    ]

    _set_labels(series, labels, rendering_kw)

    # if multiple arrows are given, plot them all in one plot
    show = kwargs.pop("show", True)
    B = _instantiate_backend(Backend, series[0], show=False, **kwargs)
    for i in range(1, len(series)):
        B.extend(_instantiate_backend(Backend, series[i], show=False, **kwargs))
    if show:
        B.show()
    return B



from spb.functions import plot_parametric
from sympy.matrices import MatrixBase
from spb.backends.base_backend import Plot


def plot_boundary(param_func: MatrixBase, *args, **kwargs) -> (Plot):
    """A function for plotting the boundary of a 2D area

    Args:
        param_func (MatrixBase,): A matrix with functions [x(u,v), y(u,v)] determining x and y-coordinates of area to be plotted.
        U_lim (Tuple[Symbol, float, float]): A tuple with variable and its limits either as list or two seperate values. Example as (u, u_min, u_max).
        V_lim (Tuple[Symbol, float, float]): A tuple with variable and its limits either as list or two seperate values. Example as (v, v_min, v_max).
        rendering_kw (dict, optional): A dictionary forwarded to dtuplot.plot(), see SPB docs for reference.
        color (str, optional): A string to set color boundary with. With no argument color = 'blue'.
        show (bool, optional): Boolean, if 'True': show plot, other just return object without plotting. Defaults to 'True'.


    Returns:
        Plot: A SPB-plot object
    """
    if len(param_func) == 2:
        bounds = []
        for i, arg in enumerate(args):
            if isinstance(arg, tuple):
                bounds.append(arg)
                if isinstance(arg[1], list):
                    bounds[-1] = (arg[0], *arg[1])
            else:
                rest = args[i + 1 :]
                args = rest
                break
        else:
            args = []
        if len(bounds) != 2:
            raise ValueError("For 2D-areas two variables are needed to plot boundary")

        show = kwargs.pop("show", True)
        color = kwargs.pop("color", "blue")
        adaptive = kwargs.setdefault("adaptive", False)
        plot = plot_parametric(show=False)
        for i in range(2):
            plot += plot_parametric(
                *param_func.subs(bounds[i][0], bounds[i][1]),
                (bounds[1 - i][0], *bounds[1 - i][1:]),
                {"color": color},
                *args,
                show=False,
                **kwargs
            )
            plot += plot_parametric(
                *param_func.subs(bounds[i][0], bounds[i][2]),
                (bounds[1 - i][0], *bounds[1 - i][1:]),
                {"color": color},
                *args,
                show=False,
                **kwargs
            )
    elif len(param_func) == 3:
        raise NotImplementedError("Volume boundary plots are not yet implemented")

    plot.legend = kwargs.pop("legend", False)
    plot.xlabel = kwargs.pop("xlabel", "x")
    plot.ylabel = kwargs.pop("ylabel", "y")
    plot.zlabel = kwargs.pop("zlabel", "z")
    plot.aspect = kwargs.pop("aspect", "auto")

    if show:
        plot.show()
        return plot
    else:
        return plot



# Following are the helper functions such that data can be read
# and passed on correctly for all backends

# Matplotlib
from spb.backends.matplotlib.renderers.vector2d import (
    _draw_vector2d_helper as MB_draw_vector2d_helper,
    _update_vector2d_helper as MB_update_vector2d_helper,
)
from spb.backends.matplotlib.renderers.vector3d import (
    _draw_vector3d_helper as MB_draw_vector3d_helper,
    _update_vector3d_helper as MB_update_vector3d_helper,
)
from spb.backends.matplotlib.renderers.renderer import MatplotlibRenderer


def MB_draw_quiver2d_helper(renderer, data):
    start, end = data.T
    direction = end - start
    xx, yy = start
    uu, vv = direction
    return MB_draw_vector2d_helper(renderer, (xx, yy, uu, vv))


def MB_update_quiver2d_helper(renderer, data, handle):
    start, end = data.T
    direction = end - start
    xx, yy = start
    uu, vv = direction
    return MB_update_vector2d_helper(renderer, (xx, yy, uu, vv), handle)


class MBQuiver2DRenderer(MatplotlibRenderer):
    draw_update_map = {MB_draw_quiver2d_helper: MB_update_quiver2d_helper}
    
    def __init__(self, plot, s):
        super().__init__(plot,s)
        # set "_sal" based off series
        self._sal = s._sal


def MB_draw_quiver3d_helper(renderer, data):
    start, end = data.T
    direction = end - start
    xx, yy, zz = start
    uu, vv, ww = direction
    return MB_draw_vector3d_helper(renderer, (xx, yy, zz, uu, vv, ww))


def MB_update_quiver3d_helper(renderer, data, handle):
    start, end = data.T
    direction = end - start
    xx, yy, zz = start
    uu, vv, ww = direction
    return MB_update_vector3d_helper(renderer, (xx, yy, zz, uu, vv, ww), handle)


class MBQuiver3DRenderer(MatplotlibRenderer):
    draw_update_map = {MB_draw_quiver3d_helper: MB_update_quiver3d_helper}
    
    def __init__(self, plot, s):
        super().__init__(plot,s)
        # set "_sal" based off series
        self._sal = s._sal


# Plotly
from spb.backends.plotly.renderers.vector2d import (
    _draw_vector2d_helper as PB_draw_vector2d_helper,
    _update_vector2d_helper as PB_update_vector2d_helper,
)
from spb.backends.plotly.renderers.vector3d import (
    _draw_vector3d_helper as PB_draw_vector3d_helper,
    _update_vector3d_helper as PB_update_vector3d_helper,
)
from spb.backends.base_renderer import Renderer


def PB_draw_quiver2d_helper(renderer, data):
    start, end = data.T
    direction = end - start
    xx, yy = start
    uu, vv = direction
    return PB_draw_vector2d_helper(renderer, ([xx], [yy], [uu], [vv]))


def PB_update_quiver2d_helper(renderer, data, handle):
    start, end = data.T
    direction = end - start
    xx, yy = start
    uu, vv = direction
    return PB_update_vector2d_helper(renderer, ([xx], [yy], [uu], [vv]), handle)


class PBQuiver2DRenderer(Renderer):
    draw_update_map = {PB_draw_quiver2d_helper: PB_update_quiver2d_helper}


def PB_draw_quiver3d_helper(renderer, data):
    start, end = data.T
    direction = end - start
    xx, yy, zz = start
    uu, vv, ww = direction
    return PB_draw_vector3d_helper(renderer, (xx, yy, zz, uu, vv, ww))


def PB_update_quiver3d_helper(renderer, data, handle):
    start, end = data.T
    direction = end - start
    xx, yy, zz = start
    uu, vv, ww = direction
    return PB_update_vector3d_helper(renderer, (xx, yy, zz, uu, vv, ww), handle)


class PBQuiver3DRenderer(Renderer):
    draw_update_map = {PB_draw_quiver3d_helper: PB_update_quiver3d_helper}


# Bokeh
from spb.backends.bokeh.renderers.vector2d import (
    _draw_vector2d_helper as BB_draw_vector2d_helper,
    _update_vector2d_helper as BB_update_vector2d_helper,
)


def BB_draw_quiver2d_helper(renderer, data):
    start, end = data.T
    direction = end - start
    xx, yy = start
    uu, vv = direction
    return BB_draw_vector2d_helper(renderer, (xx, yy, uu, vv))


def BB_update_quiver2d_helper(renderer, data, handle):
    start, end = data.T
    direction = end - start
    xx, yy = start
    uu, vv = direction
    return BB_update_vector2d_helper(renderer, (xx, yy, uu, vv), handle)


class BBQuiver2DRenderer(Renderer):
    draw_update_map = {BB_draw_quiver2d_helper: BB_update_quiver2d_helper}


# K3D
from spb.backends.k3d.renderers.vector3d import (
    _draw_vector3d_helper as KB_draw_vector3d_helper,
    _update_vector3d_helper as KB_update_vector3d_helper,
)


def KB_draw_quiver3d_helper(renderer, data):
    start, end = data.T
    direction = end - start
    xx, yy, zz = start
    uu, vv, ww = direction
    return KB_draw_vector3d_helper(renderer, (xx, yy, zz, uu, vv, ww))


def KB_update_quiver3d_helper(renderer, data, handle):
    start, end = data.T
    direction = end - start
    xx, yy, zz = start
    uu, vv, ww = direction
    return KB_update_vector3d_helper(renderer, (xx, yy, zz, uu, vv, ww), handle)


class KBQuiver3DRenderer(Renderer):
    draw_update_map = {KB_draw_quiver3d_helper: KB_update_quiver3d_helper}


# Mayavi
from spb.backends.mayavi.renderers.vector3d import (
    _draw_vector3d_helper as MAB_draw_vector3d_helper,
    _update_vector3d_helper as MAB_update_vector3d_helper,
)


def MAB_draw_quiver3d_helper(renderer, data):
    start, end = data.T
    direction = end - start
    xx, yy, zz = start
    uu, vv, ww = direction
    return MAB_draw_vector3d_helper(renderer, (xx, yy, zz, uu, vv, ww))


def MAB_update_quiver3d_helper(renderer, data, handle):
    start, end = data.T
    direction = end - start
    xx, yy, zz = start
    uu, vv, ww = direction
    return MAB_update_vector3d_helper(renderer, (xx, yy, zz, uu, vv, ww), handle)


class MABQuiver3DRenderer(Renderer):
    draw_update_map = {MAB_draw_quiver3d_helper: MAB_update_quiver3d_helper}



from spb import *
from .quiverplot import quiver, Arrow2DSeries, Arrow3DSeries
from .scatterplot import scatter
from .boundaryplot import plot_boundary
from spb.defaults import get_default_settings, cfg, set_defaults

# Backends to support the vector plotting functionality.
from .quiverplot_helpers import (
    MBQuiver2DRenderer as MBV2,
    MBQuiver3DRenderer as MBV3,
    PBQuiver2DRenderer as PBV2,
    PBQuiver3DRenderer as PBV3,
    BBQuiver2DRenderer as BBV2,
    KBQuiver3DRenderer as KBV3,
    MABQuiver3DRenderer as MABV3
)

MB.renderers_map.update({
    Arrow2DSeries: MBV2,
    Arrow3DSeries: MBV3,
})
PB.renderers_map.update({
    Arrow2DSeries: PBV2,
    Arrow3DSeries: PBV3,
})
BB.renderers_map.update({
    Arrow2DSeries: BBV2,
})
KB.renderers_map.update({
    Arrow3DSeries: KBV3,
})
MAB.renderers_map.update({
    Arrow3DSeries: MABV3
})

# Update Mayavi functionality to support scatter plotting
from .scatterplot import Point3DRenderer
from spb.series import List3DSeries
MAB.renderers_map.update({
    List3DSeries: Point3DRenderer
})


from dtumathtools import *
from sympy import *
from spb import MB, PB, BB, KB, MAB
import pytest
import numpy as np
import re

# remove sympy variable named "test"
test = 0
del test

# Disable mayavi tests for github actions, as this fails
# Reactivate when support for mayavi should return for 4.8.2
test_mab = False

def test_quiver():
    # matplotlib
    dtuplot.quiver(
        Matrix([1, 2, 3]), Matrix([4, 5, 6]), {"color": "red"}, backend=MB, show=False
    )
    dtuplot.quiver([1, 2, 3], [4, 5, 6], backend=MB, show=False)
    dtuplot.quiver([1, 2], [4, 5], backend=MB, show=False)
    dtuplot.quiver(Matrix([1, 2]), Matrix([4, 5]), backend=MB, show=False)
    dtuplot.quiver(
        1, 2, 0, 0, 0, 3, rendering_kw={"color": "orange"}, backend=MB, show=False
    )
    dtuplot.quiver(1, 2, 1, 2, backend=MB, show=False)
    dtuplot.quiver((1, 2), (1, 2), backend=MB, show=False)
    dtuplot.quiver(np.array([1, 2]), (1, 2), backend=MB, show=False)
    dtuplot.quiver(np.array([1, 2]), np.array([1, 2]), backend=MB, show=False)
    dtuplot.quiver(np.array([1, 2]), Matrix([1, 2]), backend=MB, show=False)
    dtuplot.quiver(np.array([1, 2, 3]), (1, 2, 3), backend=MB, show=False)
    dtuplot.quiver(np.array([1, 2, 3]), np.array([1, 2, 3]), backend=MB, show=False)
    dtuplot.quiver(np.array([1, 2, 3]), Matrix([1, 2, 3]), backend=MB, show=False)
    dtuplot.quiver([1, 2, 3], Matrix([1, 2, 3]), label="123", backend=MB, show=False)
    dtuplot.quiver([1, 2, 3], Matrix([1, 2, 3]), label=["123"], backend=MB, show=False)
    dtuplot.quiver([1, 2, 3], [4, 5, 6], backend=MB, qlim=False, show=False)
    dtuplot.quiver([1, 2], [4, 5], backend=MB, qlim=False, show=False)
    # plotly
    dtuplot.quiver(Matrix([1, 2, 3]), Matrix([4, 5, 6]), show=False, backend=PB)
    dtuplot.quiver([1, 2, 3], [4, 5, 6], backend=PB, show=False)
    dtuplot.quiver([1, 2], [4, 5], backend=PB, show=False)
    dtuplot.quiver(Matrix([1, 2]), Matrix([4, 5]), backend=PB, show=False)
    dtuplot.quiver(1, 2, 0, 0, 0, 3, backend=PB, show=False)
    dtuplot.quiver(1, 2, 1, 2, backend=PB, show=False)
    dtuplot.quiver((1, 2), (1, 2), backend=PB, show=False)
    dtuplot.quiver(np.array([1, 2]), (1, 2), backend=PB, show=False)
    dtuplot.quiver(np.array([1, 2]), np.array([1, 2]), backend=PB, show=False)
    dtuplot.quiver(np.array([1, 2]), Matrix([1, 2]), backend=PB, show=False)
    dtuplot.quiver(np.array([1, 2, 3]), (1, 2, 3), backend=PB, show=False)
    dtuplot.quiver(np.array([1, 2, 3]), np.array([1, 2, 3]), backend=PB, show=False)
    dtuplot.quiver(np.array([1, 2, 3]), Matrix([1, 2, 3]), backend=PB, show=False)
    dtuplot.quiver([1, 2, 3], Matrix([1, 2, 3]), label="123", backend=PB, show=False)
    dtuplot.quiver([1, 2, 3], Matrix([1, 2, 3]), label=["123"], backend=PB, show=False)
    dtuplot.quiver([1, 2, 3], [4, 5, 6], backend=PB, qlim=False, show=False)
    dtuplot.quiver([1, 2], [4, 5], backend=PB, qlim=False, show=False)
    # bokeh
    dtuplot.quiver([1, 2], [4, 5], backend=BB, show=False)
    dtuplot.quiver(Matrix([1, 2]), Matrix([4, 5]), backend=BB, show=False)
    dtuplot.quiver(1, 2, 1, 2, backend=BB, show=False)
    dtuplot.quiver((1, 2), (1, 2), backend=BB, show=False)
    dtuplot.quiver(np.array([1, 2]), (1, 2), backend=BB, show=False)
    dtuplot.quiver(np.array([1, 2]), np.array([1, 2]), backend=BB, show=False)
    dtuplot.quiver(np.array([1, 2]), Matrix([1, 2]), backend=BB, show=False)
    dtuplot.quiver([1, 2], [4, 5], backend=BB, qlim=False, show=False)
    # k3d
    with pytest.warns(
        UserWarning, match="K3DBackend only works properly within Jupyter Notebook"
    ):
        dtuplot.quiver(Matrix([1, 2, 3]), Matrix([4, 5, 6]), backend=KB, show=False)
        dtuplot.quiver([1, 2, 3], [4, 5, 6], backend=KB, show=False)
        dtuplot.quiver(1, 2, 0, 0, 0, 3, backend=KB, show=False)
        dtuplot.quiver(np.array([1, 2, 3]), (1, 2, 3), backend=KB, show=False)
        dtuplot.quiver(np.array([1, 2, 3]), np.array([1, 2, 3]), backend=KB, show=False)
        dtuplot.quiver(np.array([1, 2, 3]), Matrix([1, 2, 3]), backend=KB, show=False)
        dtuplot.quiver(
            [1, 2, 3], Matrix([1, 2, 3]), label="123", backend=KB, show=False
        )
        dtuplot.quiver(
            [1, 2, 3], Matrix([1, 2, 3]), label=["123"], backend=KB, show=False
        )
        dtuplot.quiver([1, 2, 3], [4, 5, 6], backend=KB, qlim=False, show=False)
    # mayavi
    if test_mab:
        # A lot of things in the Mayavi toolbox is deprecated...
        with pytest.warns(DeprecationWarning):
            # # Options to hide screen popping up (commented out for github actions testing)
            # from mayavi import mlab  # used for the mayavi test to disable popups
            # mlab.options.offscreen = True
            dtuplot.quiver(
                Matrix([1, 2, 3]), Matrix([4, 5, 6]), backend=MAB, show=False, warning=False
            )
            dtuplot.quiver([1, 2, 3], [4, 5, 6], backend=MAB, show=False, warning=False)
            dtuplot.quiver(1, 2, 0, 0, 0, 3, backend=MAB, show=False, warning=False)
            dtuplot.quiver(
                np.array([1, 2, 3]), (1, 2, 3), backend=MAB, show=False, warning=False
            )
            dtuplot.quiver(
                np.array([1, 2, 3]),
                np.array([1, 2, 3]),
                backend=MAB,
                show=False,
                warning=False,
            )
            dtuplot.quiver(
                np.array([1, 2, 3]),
                Matrix([1, 2, 3]),
                backend=MAB,
                show=False,
                warning=False,
            )
            dtuplot.quiver(
                [1, 2, 3],
                Matrix([1, 2, 3]),
                label="123",
                backend=MAB,
                show=False,
                warning=False,
            )
            dtuplot.quiver(
                [1, 2, 3],
                Matrix([1, 2, 3]),
                label=["123"],
                backend=MAB,
                show=False,
                warning=False,
            )
            dtuplot.quiver(
                [1, 2, 3],
                [4, 5, 6],
                backend=MAB,
                qlim=False,
                show=False,
                warning=False
            )

        with pytest.raises(
            NotImplementedError, match="Mayavi backend does not support 2D vector plots!"
        ):
            dtuplot.quiver([1, 2], [4, 5], backend=MAB, show=False, warning=False)

        with pytest.warns(
            UserWarning,
            match="Because of the Mayavi backend, the origin of the vector might be slightly off. To supress this warning, set 'warning=False'",
        ):
            dtuplot.quiver(
                Matrix([100, 100, 100]), Matrix([10, 20, 30]), backend=dtuplot.MAB
            )

    with pytest.raises(
        NotImplementedError, match="K3D backend does not support 2D vector plots!"
    ):
        dtuplot.quiver(Matrix([1, 2]), Matrix([0, 10]), backend=dtuplot.KB)

    with pytest.raises(
        NotImplementedError, match="Bokeh backend does not support 3D vector plots!"
    ):
        dtuplot.quiver(Matrix([1, 2, 3]), Matrix([0, 10, 100]), backend=dtuplot.BB)

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Error! Wrong format used in quiver. Got [1. 2. 3. 4.] as starting point(s) and [4. 5. 6. 7.] as ending point(s)!"
        ),
    ):
        dtuplot.quiver(
            Matrix([1, 2, 3, 4]), Matrix([4, 5, 6, 7]), backend=MB, show=False
        )

    with pytest.raises(
        Exception,
        match="Error! Wrong format used in quiver. Got 7 arguments that could be start or direction vector coordinates!",
    ):
        dtuplot.quiver(1, 2, 3, 4, 5, 6, 7, backend=MB, show=False)

    with pytest.raises(
        Exception,
        match="Error! Wrong format used in quiver. Got 5 arguments that could be start or direction vector coordinates!",
    ):
        dtuplot.quiver(1, 2, 3, 4, 5, backend=MB, show=False)

    with pytest.raises(
        Exception,
        match="Error! Wrong format used in quiver. Got 3 arguments that could be start or direction vector coordinates!",
    ):
        dtuplot.quiver(1, 2, 3, backend=MB, show=False)

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Error! Start and direction vectors must be provided (only two vectors)!"
        ),
    ):
        dtuplot.quiver([1, 2], [3, 4], [5, 6], backend=MB, show=False)

    with pytest.raises(
        AssertionError,
        match=re.escape(
            "Error! Start and direction vectors must be provided (only two vectors)!"
        ),
    ):
        dtuplot.quiver([1, 2], backend=MB, show=False)

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Error! Wrong format used in quiver. Got [1.] as starting point(s) and [2.] as ending point(s)!"
        ),
    ):
        dtuplot.quiver([1], [2], backend=MB, show=False)

    with pytest.raises(
        AssertionError,
        match="Error! Number of labels must be equal to number of arrows, or empty list!",
    ):
        dtuplot.quiver(
            [1, 2, 3], Matrix([1, 2, 3]), label=["123", "456"], backend=MB, show=False
        )

    with pytest.raises(
        AssertionError, match="Error! Label must be a list or a string!"
    ):
        dtuplot.quiver([1, 2, 3], Matrix([1, 2, 3]), label=123, backend=MB, show=False)

    with pytest.raises(
        NotImplementedError, match="Interactive quiver plots are not yet implemented!"
    ):
        dtuplot.quiver([1, 2, 3], Matrix([1, 2, 3]), params=123, backend=MB, show=False)


def test_boundary():
    x, y, u, v = symbols("x y u v")
    param_func = Matrix([u, (1 - abs(u)) ** 2 * v])
    ulim = (u, -1.5, 1.5)
    vlim = (v, 0.5, 1)

    p0 = dtuplot.plot_boundary(
        param_func,
        ulim,
        vlim,
        rendering_kw={"color": "red", "linewidth": 5},
        xlabel="some",
        ylabel="thing",
        zlabel="here",
        backend=MB,
        show=False,
    )

    assert type(p0) == MB
    assert p0._series[0].is_point == False
    assert p0._series[0].is_polar == False
    assert p0._series[0].is_filled == True
    assert p0._series[0].only_integers == False
    assert p0._series[0].show_in_legend == True
    assert p0._series[0].adaptive == False
    assert p0._series[0]._scales == ["linear", "linear", "linear"]
    assert p0._series[0]._label == "v"
    assert p0._series[0].expr_x == -1.5
    assert p0._series[0].expr_y == 0.25 * v
    assert p0._series[0].var == v
    assert p0._series[0].start == 0.5
    assert p0._series[0].end == 1.0

    # testing errors
    with pytest.raises(
        ValueError, match="For 2D-areas two variables are needed to plot boundary"
    ):
        w = symbols("w)")
        wlim = (w, -1, 1)
        p0 = dtuplot.plot_boundary(
            param_func,
            ulim,
            vlim,
            wlim,
            rendering_kw={"color": "red", "linewidth": 5},
            xlabel="some",
            ylabel="thing",
            zlabel="here",
            show=False,
            backend=MB,
        )

    with pytest.raises(
        NotImplementedError, match="Volume boundary plots are not yet implemented"
    ):
        param_func = Matrix([u, (1 - abs(u)) ** 2 * v, v])  # *u/abs(u)

        p0 = dtuplot.plot_boundary(
            param_func,
            ulim,
            vlim,
            rendering_kw={"color": "red", "linewidth": 5},
            xlabel="some",
            ylabel="thing",
            zlabel="here",
            show=False,
            backend=MB,
        )


def test_scatterplot():
    # matplotlib
    dtuplot.scatter(1, 2, backend=MB, show=False)
    dtuplot.scatter(1, 2, 3, backend=MB, show=False)
    dtuplot.scatter([1], [2], backend=MB, show=False)
    dtuplot.scatter([[1]], [[2]], backend=MB, show=False)
    dtuplot.scatter([1, 2, 3], [4, 5, 6], backend=MB, show=False)
    dtuplot.scatter([1, 2, 3], [4, 5, 6], [7, 8, 9], backend=MB, show=False)
    dtuplot.scatter(
        np.array([1, 2, 3]),
        np.array([4, 5, 6]),
        np.array([7, 8, 9]),
        backend=MB,
        show=False,
    )
    dtuplot.scatter(np.array([1]), np.array([4]), np.array([7]), backend=MB, show=False)
    dtuplot.scatter(np.array([1, 2, 3]), np.array([4, 5, 6]), backend=MB, show=False)
    with pytest.warns(match="np.array with more than 1 dimension"):
        dtuplot.scatter(
            np.array([[1, 2, 3]]), np.array([[4, 5, 6]]), backend=MB, show=False
        )
    with pytest.warns(match="np.array with more than 1 dimension"):
        dtuplot.scatter(
            np.array([[1, 2, 3], [7, 8, 9]]),
            np.array([[4, 5, 6], [4, 5, 6]]),
            backend=MB,
            show=False,
        )
    dtuplot.scatter(np.array([1]), np.array([4]), backend=MB, show=False)
    dtuplot.scatter(Matrix([1, 2, 3]), Matrix([4, 5, 6]), backend=MB, show=False)
    dtuplot.scatter(Matrix([1, 2]), Matrix([4, 5]), backend=MB, show=False)

    dtuplot.scatter(Matrix([1, 2]).T, Matrix([4, 5]), backend=MB, show=False)
    dtuplot.scatter(np.array([1, 2]), Matrix([4, 5]), backend=MB, show=False)
    dtuplot.scatter(np.array([1, 2]), Matrix([4, 5]), [7, 8], backend=MB, show=False)
    dtuplot.scatter([1, 2], backend=MB, show=False)
    dtuplot.scatter((1, 2), backend=MB, show=False)
    dtuplot.scatter(Matrix([1, 2]), backend=MB, show=False)
    dtuplot.scatter([pi, 2 * pi], backend=MB, show=False)
    dtuplot.scatter(np.array([pi, 2 * pi]), backend=MB, show=False)
    dtuplot.scatter(Matrix([pi, 2 * pi]), backend=MB, show=False)
    dtuplot.scatter(Matrix([-1, 0, 1]), backend=MB, show=False)
    dtuplot.scatter(Matrix([0, 0, 0]), backend=MB, show=False)
    dtuplot.scatter(np.array([1, 2]), backend=MB, show=False)
    dtuplot.scatter([1, 2, 3], backend=MB, show=False)
    dtuplot.scatter((1, 2, 3), backend=MB, show=False)
    dtuplot.scatter(Matrix([1, 2, 3]), backend=MB, show=False)
    dtuplot.scatter(np.array([1, 2, 3]), backend=MB, show=False)
    dtuplot.scatter([[1, 2], [4, 5]], backend=MB, show=False)
    dtuplot.scatter([(1, 2), [4, 5]], backend=MB, show=False)
    dtuplot.scatter([Matrix([1, 2]), [4, 5]], backend=MB, show=False)
    dtuplot.scatter([Matrix([1, 2]), Matrix([4, 5])], backend=MB, show=False)
    dtuplot.scatter([np.array([1, 2]), [4, 5]], backend=MB, show=False)
    dtuplot.scatter([np.array([1, 2]), np.array([4, 5])], backend=MB, show=False)
    dtuplot.scatter(
        [Matrix([1, 2]), np.array([4, 5]), (7, 8), [3, 9]], backend=MB, show=False
    )
    dtuplot.scatter([[1, 2, 3], [4, 5, 6]], backend=MB, show=False)
    dtuplot.scatter([(1, 2, 3), [4, 5, 6]], backend=MB, show=False)
    dtuplot.scatter([Matrix([1, 2, 3]), [4, 5, 6]], backend=MB, show=False)
    dtuplot.scatter([Matrix([1, 2, 3]), Matrix([4, 5, 6])], backend=MB, show=False)
    dtuplot.scatter([np.array([1, 2, 3]), [4, 5, 6]], backend=MB, show=False)
    dtuplot.scatter([np.array([1, 2, 3]), np.array([4, 5, 6])], backend=MB, show=False)
    dtuplot.scatter(
        [Matrix([1, 2, 3]), [4, 5, 6], (7, 8, 9), np.array([-2, -1, 0])],
        backend=MB,
        show=False,
    )

    with pytest.raises(
        AssertionError, match="scatterplot only supports 2D and 3D plots"
    ):
        dtuplot.scatter(1, 2, 3, 4, backend=MB, show=False)
    with pytest.raises(
        AssertionError, match="scatterplot only supports 2D and 3D plots"
    ):
        dtuplot.scatter(1, backend=MB, show=False)
    # plotly
    dtuplot.scatter(1, 2, backend=PB, show=False)
    dtuplot.scatter(1, 2, 3, backend=PB, show=False)
    dtuplot.scatter([1], [2], backend=PB, show=False)
    dtuplot.scatter([[1]], [[2]], backend=PB, show=False)
    dtuplot.scatter([1, 2, 3], [4, 5, 6], backend=PB, show=False)
    dtuplot.scatter([1, 2, 3], [4, 5, 6], [7, 8, 9], backend=PB, show=False)
    dtuplot.scatter(
        np.array([1, 2, 3]),
        np.array([4, 5, 6]),
        np.array([7, 8, 9]),
        backend=PB,
        show=False,
    )
    dtuplot.scatter(np.array([1]), np.array([4]), np.array([7]), backend=PB, show=False)
    dtuplot.scatter(np.array([1, 2, 3]), np.array([4, 5, 6]), backend=PB, show=False)
    with pytest.warns(match="np.array with more than 1 dimension"):
        dtuplot.scatter(
            np.array([[1, 2, 3]]), np.array([[4, 5, 6]]), backend=PB, show=False
        )
    with pytest.warns(match="np.array with more than 1 dimension"):
        dtuplot.scatter(
            np.array([[1, 2, 3], [7, 8, 9]]),
            np.array([[4, 5, 6], [4, 5, 6]]),
            backend=PB,
            show=False,
        )
    dtuplot.scatter(np.array([1]), np.array([4]), backend=PB, show=False)
    dtuplot.scatter(
        Matrix([1, 2, 3]), Matrix([4, 5, 6]), backend=PB, show=False
    )  # different behavior to the np.array version
    dtuplot.scatter(Matrix([1, 2]), Matrix([4, 5]), backend=PB, show=False)
    with pytest.raises(
        AssertionError, match="scatterplot only supports 2D and 3D plots"
    ):
        dtuplot.scatter(1, 2, 3, 4, backend=PB, show=False)
    with pytest.raises(
        AssertionError, match="scatterplot only supports 2D and 3D plots"
    ):
        dtuplot.scatter(1, backend=PB, show=False)
    dtuplot.scatter(Matrix([1, 2]).T, Matrix([4, 5]), backend=PB, show=False)
    dtuplot.scatter(np.array([1, 2]), Matrix([4, 5]), backend=PB, show=False)
    dtuplot.scatter(np.array([1, 2]), Matrix([4, 5]), [7, 8], backend=PB, show=False)
    dtuplot.scatter([1, 2], backend=PB, show=False)
    dtuplot.scatter((1, 2), backend=PB, show=False)
    dtuplot.scatter(Matrix([1, 2]), backend=PB, show=False)
    dtuplot.scatter(Matrix([-1, 0, 1]), backend=PB, show=False)
    dtuplot.scatter(Matrix([0, 0, 0]), backend=PB, show=False)
    dtuplot.scatter(np.array([1, 2]), backend=PB, show=False)
    dtuplot.scatter([1, 2, 3], backend=PB, show=False)
    dtuplot.scatter((1, 2, 3), backend=PB, show=False)
    dtuplot.scatter(Matrix([1, 2, 3]), backend=PB, show=False)
    dtuplot.scatter(np.array([1, 2, 3]), backend=PB, show=False)
    dtuplot.scatter([[1, 2], [4, 5]], backend=PB, show=False)
    dtuplot.scatter([(1, 2), [4, 5]], backend=PB, show=False)
    dtuplot.scatter([Matrix([1, 2]), [4, 5]], backend=PB, show=False)
    dtuplot.scatter([Matrix([1, 2]), Matrix([4, 5])], backend=PB, show=False)
    dtuplot.scatter([np.array([1, 2]), [4, 5]], backend=PB, show=False)
    dtuplot.scatter([np.array([1, 2]), np.array([4, 5])], backend=PB, show=False)
    dtuplot.scatter(
        [Matrix([1, 2]), np.array([4, 5]), (7, 8), [3, 9]], backend=PB, show=False
    )
    dtuplot.scatter([[1, 2, 3], [4, 5, 6]], backend=PB, show=False)
    dtuplot.scatter([(1, 2, 3), [4, 5, 6]], backend=PB, show=False)
    dtuplot.scatter([Matrix([1, 2, 3]), [4, 5, 6]], backend=PB, show=False)
    dtuplot.scatter([Matrix([1, 2, 3]), Matrix([4, 5, 6])], backend=PB, show=False)
    dtuplot.scatter([np.array([1, 2, 3]), [4, 5, 6]], backend=PB, show=False)
    dtuplot.scatter([np.array([1, 2, 3]), np.array([4, 5, 6])], backend=PB, show=False)
    dtuplot.scatter(
        [Matrix([1, 2, 3]), [4, 5, 6], (7, 8, 9), np.array([-2, -1, 0])],
        backend=PB,
        show=False,
    )
    # bokeh
    dtuplot.scatter(1, 2, backend=BB, show=False)
    dtuplot.scatter([1], [2], backend=BB, show=False)
    dtuplot.scatter([[1]], [[2]], backend=BB, show=False)
    dtuplot.scatter([1, 2, 3], [4, 5, 6], backend=BB, show=False)
    dtuplot.scatter(np.array([1, 2, 3]), np.array([4, 5, 6]), backend=BB, show=False)
    dtuplot.scatter(np.array([1]), np.array([4]), backend=BB, show=False)
    dtuplot.scatter(Matrix([1, 2]), Matrix([4, 5]), backend=BB, show=False)
    dtuplot.scatter(Matrix([1, 2]).T, Matrix([4, 5]), backend=BB, show=False)
    dtuplot.scatter(np.array([1, 2]), Matrix([4, 5]), backend=BB, show=False)
    dtuplot.scatter(np.array([1, 2]), [7, 8], backend=BB, show=False)
    dtuplot.scatter([1, 2], backend=BB, show=False)
    dtuplot.scatter((1, 2), backend=BB, show=False)
    dtuplot.scatter(Matrix([1, 2]), backend=BB, show=False)
    dtuplot.scatter(Matrix([1.0, 2.0]), backend=BB, show=False)
    dtuplot.scatter(Matrix([-1, 0]), backend=BB, show=False)
    dtuplot.scatter(np.array([1, 2]), backend=BB, show=False)
    dtuplot.scatter([[1, 2], [4, 5]], backend=BB, show=False)
    dtuplot.scatter([(1, 2), [4, 5]], backend=BB, show=False)
    dtuplot.scatter([Matrix([1, 2]), [4, 5]], backend=BB, show=False)
    dtuplot.scatter([Matrix([1, 2]), Matrix([4, 5])], backend=BB, show=False)
    dtuplot.scatter([np.array([1, 2]), [4, 5]], backend=BB, show=False)
    dtuplot.scatter([np.array([1, 2]), np.array([4, 5])], backend=BB, show=False)
    dtuplot.scatter(
        [Matrix([1, 2]), np.array([4, 5]), (7, 8), [3, 9]], backend=BB, show=False
    )
    # k3d
    with pytest.warns(
        UserWarning, match="K3DBackend only works properly within Jupyter Notebook"
    ):
        dtuplot.scatter(1, 2, 3, backend=KB, show=False)
        dtuplot.scatter([1, 2, 3], [4, 5, 6], [7, 8, 9], backend=KB, show=False)
        dtuplot.scatter(
            np.array([1, 2, 3]),
            np.array([4, 5, 6]),
            np.array([7, 8, 9]),
            backend=KB,
            show=False,
        )
        dtuplot.scatter(
            np.array([1]), np.array([4]), np.array([7]), backend=KB, show=False
        )
        dtuplot.scatter(
            Matrix([1, 2, 3]).T,
            Matrix([4, 5, 6]),
            Matrix([[4], [5], [6]]),
            backend=KB,
            show=False,
        )
        dtuplot.scatter(
            np.array([1, 2, 3]), Matrix([4, 5, 6]), [7, 8, 9], backend=KB, show=False
        )
        dtuplot.scatter(Matrix([-1, 0, 1]), backend=KB, show=False)
        dtuplot.scatter(Matrix([0, 0, 0]), backend=KB, show=False)
        dtuplot.scatter([1, 2, 3], backend=KB, show=False)
        dtuplot.scatter((1, 2, 3), backend=KB, show=False)
        dtuplot.scatter(Matrix([1, 2, 3]), backend=KB, show=False)
        dtuplot.scatter(np.array([1, 2, 3]), backend=KB, show=False)
        dtuplot.scatter([[1, 2, 3], [4, 5, 6]], backend=KB, show=False)
        dtuplot.scatter([(1, 2, 3), [4, 5, 6]], backend=KB, show=False)
        dtuplot.scatter([Matrix([1, 2, 3]), [4, 5, 6]], backend=KB, show=False)
        dtuplot.scatter([Matrix([1, 2, 3]), Matrix([4, 5, 6])], backend=KB, show=False)
        dtuplot.scatter([np.array([1, 2, 3]), [4, 5, 6]], backend=KB, show=False)
        dtuplot.scatter(
            [np.array([1, 2, 3]), np.array([4, 5, 6])], backend=KB, show=False
        )
        dtuplot.scatter(
            [Matrix([1, 2, 3]), [4, 5, 6], (7, 8, 9), np.array([-2, -1, 0])],
            backend=KB,
            show=False,
        )
    # mayavi
    if test_mab:
        with pytest.warns(DeprecationWarning):
            dtuplot.scatter(1, 2, 3, backend=MAB, show=False)
            dtuplot.scatter([1, 2, 3], [4, 5, 6], [7, 8, 9], backend=MAB, show=False)
            dtuplot.scatter(
                np.array([1, 2, 3]),
                np.array([4, 5, 6]),
                np.array([7, 8, 9]),
                backend=MAB,
                show=False,
            )
            dtuplot.scatter(
                np.array([1]), np.array([4]), np.array([7]), backend=MAB, show=False
            )
            dtuplot.scatter(
                Matrix([1, 2, 3]).T,
                Matrix([4, 5, 6]),
                Matrix([[4], [5], [6]]),
                backend=MAB,
                show=False,
            )
            dtuplot.scatter(
                np.array([1, 2, 3]), Matrix([4, 5, 6]), [7, 8, 9], backend=MAB, show=False
            )
            dtuplot.scatter(Matrix([-1, 0, 1]), backend=MAB, show=False)
            dtuplot.scatter(Matrix([0, 0, 0]), backend=MAB, show=False)
            dtuplot.scatter([1, 2, 3], backend=MAB, show=False)
            dtuplot.scatter((1, 2, 3), backend=MAB, show=False)
            dtuplot.scatter(Matrix([1, 2, 3]), backend=MAB, show=False)
            dtuplot.scatter(np.array([1, 2, 3]), backend=MAB, show=False)
            dtuplot.scatter([[1, 2, 3], [4, 5, 6]], backend=MAB, show=False)
            dtuplot.scatter([(1, 2, 3), [4, 5, 6]], backend=MAB, show=False)
            dtuplot.scatter([Matrix([1, 2, 3]), [4, 5, 6]], backend=MAB, show=False)
            dtuplot.scatter([Matrix([1, 2, 3]), Matrix([4, 5, 6])], backend=MAB, show=False)
            dtuplot.scatter([np.array([1, 2, 3]), [4, 5, 6]], backend=MAB, show=False)
            dtuplot.scatter(
                [np.array([1, 2, 3]), np.array([4, 5, 6])], backend=MAB, show=False
            )
            dtuplot.scatter(
                [Matrix([1, 2, 3]), [4, 5, 6], (7, 8, 9), np.array([-2, -1, 0])],
                backend=MAB,
                show=False,
            )

    # testing arguments
    dtuplot.scatter(
        1,
        2,
        3,
        rendering_kw={"alpha": 1, "s": 100, "color": "black"},
        backend=MB,
        show=False,
    )
    dtuplot.scatter(
        1, -1, rendering_kw={"color": "black", "s": 10}, backend=MB, show=False
    )
    dtuplot.scatter(
        1,
        2,
        rendering_kw={"markersize": 10, "color": "r"},
        xlim=[-2, 2],
        ylim=[-2, 2],
        backend=MB,
        show=False,
    )

    # test assertions
    if test_mab:
        with pytest.raises(
            NotImplementedError, match="Mayavi does not support 2D scatter plots!"
        ):
            dtuplot.scatter([1, 2, 3], [4, 5, 6], backend=MAB, show=False)
    with pytest.raises(
        NotImplementedError, match="K3D does not support 2D scatter plots!"
    ):
        dtuplot.scatter([1, 2, 3], [4, 5, 6], backend=KB, show=False)
    with pytest.raises(
        NotImplementedError, match="Bokeh does not support 3D scatter plots!"
    ):
        dtuplot.scatter([1, 2, 3], [4, 5, 6], [7, 8, 9], backend=BB, show=False)

    with pytest.warns(match="non-Expr objects in a Matrix is deprecated."):
        item = Matrix([[[1, 2]]])
    with pytest.raises(
        AssertionError, match="scatterplot only supports 2D and 3D plots"
    ):
        dtuplot.scatter(item, item, item, item, backend=MB, show=False)

    with pytest.raises(ValueError, match="Unknown input found"):
        dtuplot.scatter(
            NotImplementedError,
            NotImplementedError,
            NotImplementedError,
            backend=dtuplot.MB,
            show=False,
        )

    with pytest.raises(
        AssertionError,
        match=re.escape("Points given (single or list of) must be 2D or 3D!"),
    ):
        dtuplot.scatter(np.array([1, 2, 3, 4]), backend=dtuplot.MB, show=False)

    with pytest.raises(ValueError, match="Invalid type of coordinate/point"):
        dtuplot.scatter([dtuplot], backend=dtuplot.MB, show=False)

    with pytest.raises(AssertionError, match="Invalid type of coordinate, recieved"):
        dtuplot.scatter([[dtuplot]], backend=dtuplot.MB, show=False)

    with pytest.raises(
        AssertionError, match="Length of all points in list must match!"
    ):
        dtuplot.scatter([[1, 2, 3], [1, 2]], backend=dtuplot.MB, show=False)



from sympy import MatrixBase, Expr, Basic, MutableMatrix, ImmutableMatrix
from sympy.external import import_module
from spb.functions import plot_list, plot3d_list
from spb.backends.base_backend import Plot
import numpy as np
from spb import MB, PB, BB, KB, MAB
from spb.defaults import THREE_D_B, TWO_D_B


def scatter(*args, **kwargs):
    """Create a plot with one/multiple point(s). Similar to plt.scatter.

    Args:
        points (MatrixBase, np.ndarray, list, float): The point(s) to scatter. Format is 'x, y, [z], **kwargs' with x,y,[z] being Matrix, list, float or np.ndarray. If a single Matrix, list or np.ndarray is given, each entry will be treated as seperate a point.
        rendering_kw (dict, optional): A dictionary forwarded to dtuplot.plot(), see SPB docs for reference.
        color (str, optional): A string to set the color of the points with. With no argument color = 'blue'.
        show (bool, optional): Boolean, if 'True': show plot, other just return object without plotting. Defaults to 'True'.

    Returns:
        Plot: The plot containing the points.
    """

    np = import_module("numpy")
    warnings = import_module("warnings")

    def warning_formatter(msg, *args, line=None, **kwargs):
        return str(msg)

    warnings.formatwarning = warning_formatter

    # format all arguments
    dim = 0
    args = list(args)
    if len(args) == 1 and type(args[0]) in [
        list,
        tuple,
        MutableMatrix,
        ImmutableMatrix,
        np.ndarray,
    ]:
        # Single list/tuple/matrix/array given, assume this is list of (single) point(s)
        coords = [[], [], []]
        dim = None
        for i, arg in enumerate(args[0]):
            if (
                type(arg) in [float, int, Expr]
                or np.issubdtype(type(arg), np.integer)
                or np.issubdtype(type(arg), np.floating)
                or isinstance(arg, Basic)
            ):
                # the single arg is a single point
                dim = len(args[0])
                coords[i].append(float(arg))
            elif type(arg) in [list, tuple, MutableMatrix, ImmutableMatrix, np.ndarray]:
                # Unify format
                arg = np.array(arg).flatten()
                # Check dimension
                if dim is None:
                    dim = len(arg)
                else:
                    assert len(arg) == dim, "Length of all points in list must match!"
                # Sort the coordinates into correct bins
                for o, coord in enumerate(arg):
                    assert (
                        type(coord) in [float, int, Expr]
                        or np.issubdtype(type(coord), np.integer)
                        or np.issubdtype(type(coord), np.floating)
                        or isinstance(coord, Basic)
                    ), f"Invalid type of coordinate, recieved {coord} with type {str(type(coord))}"
                    coords[o].append(float(coord))
            else:
                raise ValueError(
                    f"Invalid type of coordinate/point, recieved {arg} with type {str(type(arg))}, but must be one of [list, tuple, Matrix, np.ndarray, int, float, Expr]!"
                )
            assert dim in [2, 3], "Points given (single or list of) must be 2D or 3D!"

        if len(coords[-1]) == 0:
            assert dim == 2, "Unspecified error!"
            del coords[-1]

        args = coords
    else:
        # Probably a list of arguments given!
        for i in range(len(args)):
            if type(args[i]) in (list, tuple, MutableMatrix, ImmutableMatrix):
                # This is x, y, or z argument. List of these coordinates for each point.
                args[i] = np.array(args[i]).flatten()
                dim += 1
            elif type(args[i]) == np.ndarray:
                if len(args[i].shape) > 1:
                    warnings.warn(
                        f"np.array with more than 1 dimension ({len(args[i].shape)} dimensions was found) has been flattened!"
                    )
                args[i] = args[i].flatten()
                dim += 1
            elif type(args[i]) in [float, int]:
                # Single entries for x,y,z given.
                args[i] = np.array([args[i]])
                dim += 1
            else:
                try:
                    # symbolic expression not multiplied fully
                    args[i] = np.array([float(args[i])])
                    dim += 1
                except:
                    raise ValueError(f"Unknown input found: {args[i]}")

    assert dim in [
        2,
        3,
    ], f"scatterplot only supports 2D and 3D plots, but arguments for dimension {dim} was given."
    if dim == 2:
        Backend = kwargs.pop("backend", TWO_D_B)
        if Backend == KB:
            raise NotImplementedError("K3D does not support 2D scatter plots!")
        elif Backend == MAB:
            raise NotImplementedError("Mayavi does not support 2D scatter plots!")
        rendering_kw = kwargs.pop("rendering_kw", {})
        markersize = rendering_kw.pop("s", None)
        if markersize is not None:
            rendering_kw.setdefault("markersize", markersize)
        return plot_list(
            *args, is_point=True, rendering_kw=rendering_kw, backend=Backend, **kwargs
        )
    else:
        Backend = kwargs.pop("backend", THREE_D_B)
        if Backend == BB:
            raise NotImplementedError("Bokeh does not support 3D scatter plots!")
        if Backend == MAB:
            warnings.warn(f"Mayavi is not currently supported by dtumathtools.")
        return plot3d_list(*args, is_point=True, backend=Backend, **kwargs)


# Adjust renderer for Mayavi to support 3D scatter plotting
from spb.backends.mayavi.renderers.line3d import (
    _draw_line3d_helper,
    _update_line3d_helper,
)
from spb.backends.base_renderer import Renderer


def MAB_draw_line3d_helper(renderer, data):
    x, y, z = data
    u = np.ones_like(x)
    return _draw_line3d_helper(renderer, (x, y, z, u))


class Point3DRenderer(Renderer):
    draw_update_map = {MAB_draw_line3d_helper: _update_line3d_helper}



import pytest
from sympy import *

from dtumathtools import *

# remove sympy variable named "test"
test = 0
del test


def test_taylor():
    x, y = symbols("x y")
    f = x**3 - 3 * x**2 + y**3 - 3 * y**2

    assert dtutools.taylor(f, [x, 0, y, 0], 3) == -3 * x**2 - 3 * y**2
    assert dtutools.taylor(f, [x, 0, y, 2], 3) == -3 * x**2 + 3 * (y - 2) ** 2 - 4
    assert dtutools.taylor(f, [x, 2, y, 0], 3) == -3 * y**2 + 3 * (x - 2) ** 2 - 4


def test_gradient():
    x, y, z = symbols("x y z")
    f = x**2 * sin(y) * z

    assert dtutools.gradient(f) == Matrix(
        [2 * x * z * sin(y), x**2 * z * cos(y), x**2 * sin(y)]
    )
    assert dtutools.gradient(f, [x, z]) == Matrix([2 * x * z * sin(y), x**2 * sin(y)])


def test_div():
    x, y, z = symbols("x y z")
    V = Matrix([x * y * sin(z), x * ln(y), -x * y * z**3])

    assert dtutools.div(V, var=[x, y, z]) == -3 * x * y * z**2 + x / y + y * sin(z)

    u, v, w = symbols("u v w")
    V2 = Matrix([u, w**2, u * sin(w)])

    assert dtutools.div(V2, var=[u, v, w]) == u * cos(w) + 1


# @pytest.mark.filterwarnings("ignore")
def test_warnings():
    x, y, z = symbols("x y z")
    V = Matrix([x * y * sin(z), x * ln(y), -x * y * z**3])

    u, v, w = symbols("u v w")
    V2 = Matrix([u, w**2, u * sin(w)])

    with pytest.warns(
        UserWarning, match="Warning! Variables were not specified."
    ):  #'Warning! Variables were not specified.'
        # Trigger a warning.
        dtutools.div(V)
        dtutools.div(V2)
        dtutools.rot(V)
        dtutools.rot(V2)


def test_errors():
    x, y, z = symbols("x y z")
    u, v, w = symbols("u v w")
    a, b, c = symbols("a b c")

    V = Matrix([x * y * sin(z), x * ln(y), -x * y * z**3])
    V2 = Matrix([u, w**2, u * sin(w)])
    V3 = Matrix([u * x, sin(y), (u - y) * x])
    V4 = Matrix([1, a * c, sin(b)])
    V5 = Matrix([x * y * sin(c), x * ln(y), -x * y * c**3])

    with pytest.raises(
        AssertionError, match="ERROR! Could not guess the variables used."
    ):
        # raise errors
        dtutools.div(V3)
        dtutools.div(V4)
        dtutools.div(V5)
        dtutools.rot(V3)
        dtutools.rot(V4)
        dtutools.rot(V5)

    with pytest.raises(AssertionError, match="Exactly 3 variables must be specified!"):
        dtutools.div(V, var=[])
        dtutools.div(V, var=[x])
        dtutools.div(V, var=[x, y, z, u, v, w])
        dtutools.div(V, var=[x, y, z, u])
        dtutools.div(V2, var=[u, w])
        dtutools.rot(V, var=[])
        dtutools.rot(V, var=[x])
        dtutools.rot(V, var=[x, y, z, u, v, w])
        dtutools.rot(V, var=[x, y, z, u])
        dtutools.rot(V2, var=[u, w])


def test_rot():
    x, y, z = symbols("x y z")
    V = Matrix([x * y * sin(z), x * ln(y), -x * y * z**3])

    assert dtutools.rot(V, var=[x, y, z]) == Matrix(
        [-x * z**3, x * y * cos(z) + y * z**3, -x * sin(z) + log(y)]
    )

    u, v, w = symbols("u v w")
    V2 = Matrix([u, w**2, u * sin(w)])

    assert dtutools.rot(V2, var=[u, v, w]) == Matrix([-2 * w, -sin(w), 0])


def test_hessian():
    x, y, z = symbols("x y z")
    f = x**4 + 4 * x**2 * y**2 + y**4 - 4 * x**3 - 4 * y**3 + 2

    assert dtutools.hessian(f) == Matrix(
        [
            [12 * x**2 - 24 * x + 8 * y**2, 16 * x * y],
            [16 * x * y, 8 * x**2 + 12 * y**2 - 24 * y],
        ]
    )
